@@ -1,11 +1,11 @@
-#!/bin/bash
+#!/bin/sh
 
 ZIP_DESTDIR="${MESON_BUILD_ROOT}/_zip"
 ZIP_DIR="${MESON_BUILD_ROOT}/${UUID}"
 ZIP_FILE="${MESON_BUILD_ROOT}/${UUID}.zip"
 
-GSCHEMA_DIR="${ZIP_DESTDIR}/${DATADIR}/glib-2.0/schemas"
-LOCALE_DIR="${ZIP_DESTDIR}/${DATADIR}/locale"
+GSCHEMA_DIR="${ZIP_DESTDIR}/${GSCHEMADIR}"
+LOCALE_DIR="${ZIP_DESTDIR}/${LOCALEDIR}"
 
 # PRE-CLEAN
 rm -rf ${ZIP_DESTDIR} ${ZIP_DIR} ${ZIP_FILE}
