@@ -80,41 +80,18 @@ var Packet = class Packet {
 /**
  * Data Channel
  */
-var Channel = GObject.registerClass({
-    GTypeName: 'GSConnectCoreChannel',
-    Signals: {
-        'connected': {
-            flags: GObject.SignalFlags.RUN_FIRST
-        },
-        'disconnected': {
-            flags: GObject.SignalFlags.RUN_FIRST
-        }
-    },
-    Properties: {
-        'certificate': GObject.ParamSpec.object(
-            'certificate',
-            'TlsCertificate',
-            'The TLS Certificate for this connection',
-            GObject.ParamFlags.READABLE,
-            Gio.TlsCertificate
-        ),
-        'type': GObject.ParamSpec.string(
-            'type',
-            'Data Channel Type',
-            'The protocol this data channel uses',
-            GObject.ParamFlags.READABLE,
-            null
-        )
+var Channel = class Channel {
+
+    constructor(params) {
+        Object.assign(this, params);
     }
-}, class Channel extends GObject.Object {
 
-    _init(deviceId, type) {
-        super._init();
+    get cancellable() {
+        if (this._cancellable === undefined) {
+            this._cancellable = new Gio.Cancellable();
+        }
 
-        // We need this to lookup the certificate in GSettings
-        this.identity = { body: { deviceId: deviceId } };
-        this.service = Gio.Application.get_default();
-        this._type = type;
+        return this._cancellable;
     }
 
     get certificate() {
@@ -125,14 +102,14 @@ var Channel = GObject.registerClass({
         return null;
     }
 
-    get type() {
-        return this._type;
+    get service() {
+        return Gio.Application.get_default();
     }
 
     /**
      * Set socket options
      */
-    async _initSocket(connection) {
+    _initSocket(connection) {
         if (connection instanceof Gio.TcpConnection) {
             connection.socket.set_keepalive(true);
             connection.socket.set_option(6, 4, 10); // TCP_KEEPIDLE
@@ -155,19 +132,23 @@ var Channel = GObject.registerClass({
                 close_base_stream: false
             });
 
-            stream.read_line_async(GLib.PRIORITY_DEFAULT, null, (stream, res) => {
-                try {
-                    let data = stream.read_line_finish(res)[0];
-                    stream.close(null);
+            stream.read_line_async(
+                GLib.PRIORITY_DEFAULT,
+                this.cancellable,
+                (stream, res) => {
+                    try {
+                        let data = stream.read_line_finish_utf8(res)[0];
+                        stream.close(null);
 
-                    // Store the identity as an object property
-                    this.identity = new Packet(data.toString());
+                        // Store the identity as an object property
+                        this.identity = new Packet(data);
 
-                    resolve(connection);
-                } catch (e) {
-                    reject(e);
+                        resolve(connection);
+                    } catch (e) {
+                        reject(e);
+                    }
                 }
-            });
+            );
         });
     }
 
@@ -181,7 +162,7 @@ var Channel = GObject.registerClass({
             connection.output_stream.write_all_async(
                 `${this.service.identity}`,
                 GLib.PRIORITY_DEFAULT,
-                null,
+                this.cancellable,
                 (stream, res) => {
                     try {
                         stream.write_all_finish(res);
@@ -194,68 +175,70 @@ var Channel = GObject.registerClass({
         });
     }
 
-    /**
-     * Verify connection certificate
-     */
-    _onAcceptCertificate(connection, peer_cert, flags) {
-        log(`Authenticating ${this.identity.body.deviceId}`);
-
-        connection.disconnect(connection._acceptCertificateId);
-
-        // Bail if the deviceId is missing
-        if (this.identity.body.deviceId === undefined) {
-            logWarning('missing deviceId', connection.get_remote_address().to_string());
-            return false;
-        }
-
-        // Get (or create) the settings for this deviceId
-        let settings = new Gio.Settings({
-            settings_schema: gsconnect.gschema.lookup(gsconnect.app_id + '.Device', true),
-            path: gsconnect.settings.path + 'device/' + this.identity.body.deviceId + '/'
-        });
-        let cert_pem = settings.get_string('certificate-pem');
-
-        // Verify the connection certificate if we have one
-        if (cert_pem !== '') {
-            return Gio.TlsCertificate.new_from_pem(cert_pem, -1).is_same(peer_cert);
-        }
-
-        // Otherwise trust on first use
-        return true;
-    }
-
     /**
      * Handshake Gio.TlsConnection
      */
-    _handshake(connection) {
-        return new Promise((resolve, reject) => {
-            connection.validation_flags = 0;
-            connection.authentication_mode = 1;
-            connection._acceptCertificateId = connection.connect(
-                'accept-certificate',
-                this._onAcceptCertificate.bind(this)
-            );
+    async _authenticate(connection) {
+        // FIXME: This is a hack, error propogation needs to be fixed
+        try {
+            // Standard TLS Handshake
+            await new Promise((resolve, reject) => {
+                connection.validation_flags = Gio.TlsCertificateFlags.EXPIRED;
+                connection.authentication_mode = Gio.TlsAuthenticationMode.REQUIRED;
 
-            connection.handshake_async(
-                GLib.PRIORITY_DEFAULT,
-                null,
-                (connection, res) => {
-                    try {
-                        connection.handshake_finish(res);
-                        resolve(connection);
-                    } catch (e) {
-                        reject(e);
+                connection.handshake_async(
+                    GLib.PRIORITY_DEFAULT,
+                    this.cancellable,
+                    (connection, res) => {
+                        try {
+                            resolve(connection.handshake_finish(res));
+                        } catch (e) {
+                            reject(e);
+                        }
                     }
+                );
+            });
+
+            // Bail if deviceId is missing
+            if (!this.identity.body.hasOwnProperty('deviceId')) {
+                throw new Error('missing deviceId');
+            }
+
+            // Get a GSettings object for this deviceId
+            let settings = new Gio.Settings({
+                settings_schema: gsconnect.gschema.lookup(gsconnect.app_id + '.Device', true),
+                path: gsconnect.settings.path + 'device/' + this.identity.body.deviceId + '/'
+            });
+            let cert_pem = settings.get_string('certificate-pem');
+
+            // If we have a certificate for this deviceId, we can verify it
+            if (cert_pem !== '') {
+                let certificate = Gio.TlsCertificate.new_from_pem(cert_pem, -1);
+                let valid = certificate.is_same(connection.peer_certificate);
+
+                // This is a fraudulent certificate; notify the user
+                if (!valid) {
+                    let error = new Error();
+                    error.name = 'AuthenticationError';
+                    error.deviceName = this.identity.body.deviceName;
+                    error.deviceHost = connection.base_io_stream.get_remote_address().address.to_string();
+                    this.service.notify_error(error);
+
+                    throw error;
                 }
-            );
-        });
+            }
+
+            return connection;
+        } catch (e) {
+            logError(e);
+        }
     }
 
     /**
      * If @connection is a Gio.TcpConnection, wrap it in Gio.TlsClientConnection
      * and initiate handshake, otherwise just return it.
      */
-    async _clientEncryption(connection) {
+    _clientEncryption(connection) {
         if (connection instanceof Gio.TcpConnection) {
             connection = Gio.TlsClientConnection.new(
                 connection,
@@ -263,7 +246,7 @@ var Channel = GObject.registerClass({
             );
             connection.set_certificate(this.service.certificate);
 
-            return this._handshake(connection);
+            return this._authenticate(connection);
         } else {
             return connection;
         }
@@ -273,14 +256,20 @@ var Channel = GObject.registerClass({
      * If @connection is a Gio.TcpConnection, wrap it in Gio.TlsServerConnection
      * and initiate handshake, otherwise just return it.
      */
-    async _serverEncryption(connection) {
+    _serverEncryption(connection) {
         if (connection instanceof Gio.TcpConnection) {
             connection = Gio.TlsServerConnection.new(
                 connection,
                 this.service.certificate
             );
 
-            return this._handshake(connection);
+            // If we're the server, we trust-on-first-use and verify after
+            let _id = connection.connect('accept-certificate', (conn) => {
+                conn.disconnect(_id);
+                return true;
+            });
+
+            return this._authenticate(connection);
         } else {
             return connection;
         }
@@ -292,76 +281,52 @@ var Channel = GObject.registerClass({
      * @param {Device.Device} - The device to attach to
      */
     attach(device) {
-        // Disconnect any existing channel
+        // Detach any existing channel
         if (device._channel !== null && device._channel !== this) {
-            GObject.signal_handlers_destroy(device._channel);
+            device._channel.cancellable.disconnect(device._channel._id);
             device._channel.close();
             device._channel = null;
         }
 
+        // Attach the new channel and parse it's identity
         device._channel = this;
-
-        // Parse the channel's identity packet and connect signals to the device
+        this._id = this.cancellable.connect(device._setDisconnected.bind(device));
         device._handleIdentity(this.identity);
-        this.connect('connected', device._onConnected.bind(device));
-        this.connect('disconnected', device._onDisconnected.bind(device));
 
-        // Setup pollable streams for packet exchange
+        // Setup streams for packet exchange
         this.input_stream = new Gio.DataInputStream({
             base_stream: this._connection.input_stream
         });
 
-        this.output_stream = new Gio.DataOutputStream({
-            base_stream: this._connection.output_stream
-        });
+        this.output_stream = this._connection.output_stream;
 
-        // Attach a source to the input channel, bound to the device
-        this._monitor = this._connection.input_stream.create_source(null);
-        this._monitor.set_callback(this.receive.bind(this, device));
-        this._monitor.attach(null);
-
-        // TODO: Plugins should be reloaded if we're swapping channel types.
-        //       This is flakey in general, which may be Android's fault or
-        //       possibly inherent in the protocol.
+        // TODO: If we're swapping in a different channel type, which is already
+        // flakey, we should reload or disable plugins it doesn't support
+        //device.reloadPlugins();
 
         // Emit connected:: if necessary
         if (!device.connected) {
-            this.emit('connected');
+            device._setConnected();
         }
+
+        // Start listening for packets
+        this.receive(device);
     }
 
     /**
      * Open an outgoing connection
      *
-     * Outgoing connections are opened in response to a received (or cached) UDP
-     * packet, with a mandatory kdeconnect.identity packet being sent when the
-     * connection is accepted.
-     *
-     * @param {Gio.InetSocketAddress} address - The address to open a connection
+     * @param {Gio.SocketConnection} connection - The remote connection
+     * @return {Boolean} - %true on connected, %false otherwise
      */
-    async open(address) {
-        log(`GSConnect: Opening connection to ${address.to_string()}`);
-
+    async open(connection) {
         try {
-            this._connection = await new Promise((resolve, reject) => {
-                let client = new Gio.SocketClient();
-
-                client.connect_async(address, null, (client, res) => {
-                    try {
-                        resolve(client.connect_finish(res));
-                    } catch (e) {
-                        reject(e);
-                    }
-                });
-            });
-            this._connection = await this._initSocket(this._connection);
+            this._connection = await this._initSocket(connection);
             this._connection = await this._sendIdent(this._connection);
             this._connection = await this._serverEncryption(this._connection);
 
-            this.emit('connected');
             return true;
         } catch (e) {
-            log(`GSConnect: ${e.message}`);
             debug(e);
             this.close();
             return false;
@@ -371,48 +336,36 @@ var Channel = GObject.registerClass({
     /**
      * Accept an incoming connection
      *
-     * Incoming connections are opened in response to a sent (or cached) UDP
-     * packet, with a mandatory kdeconnect.identity packet being sent when the
-     * connection is accepted.
-     *
      * @param {Gio.TcpConnection} connection - The incoming connection
+     * @return {Boolean} - %true on connected, %false otherwise
      */
     async accept(connection) {
-        if (this.type === 'tcp') {
-            let addr = connection.get_remote_address().to_string();
-            log(`GSConnect: Accepting connection from ${addr}`);
-        } else {
-            log(`GSConnect: Accepting connection from Bluez`);
-        }
-
         try {
             this._connection = await this._initSocket(connection);
             this._connection = await this._receiveIdent(this._connection);
             this._connection = await this._clientEncryption(this._connection);
 
-            this.emit('connected');
             return true;
         } catch(e) {
-            log(`GSConnect: ${e.message}`);
-            debug(e);
+            if (this.type === 'tcp') {
+                logError(e, connection.get_remote_address().to_string());
+            } else {
+                logError(e);
+            }
+
             this.close();
             return false;
         }
     }
 
     /**
-     * Close all streams associated with this channel, silencing any errors, and
-     * emit 'disconnected::'
+     * Close all streams associated with this channel, silencing any errors
      */
     close() {
-        if (this._monitor) {
-            try {
-                this._monitor.destroy();
-            } catch (e) {
-                debug(e.message);
-            }
-        }
+        // Cancel any queued operations
+        this.cancellable.cancel();
 
+        // Close any streams
         [this._connection, this.input_stream, this.output_stream].map(stream => {
             try {
                 stream.close(null);
@@ -428,125 +381,115 @@ var Channel = GObject.registerClass({
                 debug(e.message);
             }
         }
-
-        this.emit('disconnected');
-    }
-
-    /**
-     * Send a packet to a device.
-     *
-     * TODO: Currently, we don't consider failed writes to consititute a broken
-     * connection and just log a warning. This should be investigated and tested
-     * over a period of time.
-     *
-     * @param {object} packet - An object containing the packet data
-     */
-    send(packet) {
-        debug(packet, this.identity.body.deviceName);
-
-        try {
-            packet = new Packet(packet);
-            this.output_stream.put_string(packet.toString(), null);
-        } catch (e) {
-            logWarning(e, this.identity.body.deviceName);
-        }
     }
 
     /**
      * Receive a packet from the channel and call receivePacket() on the device
+     *
+     * @param {Device.Device} device - The device which will handle the packet
      */
     receive(device) {
+        this.input_stream.read_line_async(
+            GLib.PRIORITY_DEFAULT,
+            this.cancellable,
+            (stream, res) => {
+                let data, packet;
+
+                try {
+                    // Try to read and parse a packet
+                    data = stream.read_line_finish_utf8(res)[0];
+
+                    // In case %null is returned we don't want an error thrown
+                    // when trying to parse it as a packet
+                    if (data !== null) {
+                        packet = new Packet(data);
+                        debug(packet, this.identity.body.deviceName);
+                    }
+
+                    // Queue another receive() before handling the packet
+                    this.receive(device);
+
+                    if (packet) {
+                        device.receivePacket(packet);
+                    }
+                } catch (e) {
+                    // Another operation is pending, re-queue the receive()
+                    if (e.code === Gio.IOErrorEnum.PENDING) {
+                        this.receive(device);
+
+                    // Something else went wrong; disconnect
+                    } else {
+                        debug(e, this.identity.body.deviceName);
+                        this.close();
+                    }
+                }
+            }
+        );
+    }
+
+    /**
+     * Send a packet to a device
+     *
+     * See: https://github.com/KDE/kdeconnect-kde/blob/master/core/backends/lan/landevicelink.cpp#L92-L94
+     *
+     * @param {object} packet - An dictionary of packet data
+     */
+    send(packet) {
         try {
-            let data = this.input_stream.read_line(null)[0];
-            let packet = new Packet(data.toString());
+            packet = new Packet(packet);
 
-            debug(packet, device.name);
-
-            device.receivePacket(packet);
-
-            return GLib.SOURCE_CONTINUE;
+            this.output_stream.write_all_async(
+                packet.toString(),
+                GLib.PRIORITY_DEFAULT,
+                this.cancellable,
+                (stream, res) => {
+                    try {
+                        stream.write_all_finish(res);
+                        debug(packet, this.identity.body.deviceName);
+                    } catch (e) {
+                        // Another operation is pending, re-queue the send()
+                        if (e.code === Gio.IOErrorEnum.PENDING) {
+                            debug('queued', packet.type);
+                            this.send(packet);
+                        } else {
+                            debug(e);
+                        }
+                    }
+                }
+            );
         } catch (e) {
-            debug(e, this.identity.body.deviceName);
-            this.close();
-            return GLib.SOURCE_REMOVE;
+            logError(e, 'Malformed packet');
         }
     }
-});
+}
 
 
 /**
- * File Transfers
- *
- * NOTE: transfers are always closed on completion
- *
- * Example Contruction:
- *  let transfer = new Protocol.Transfer({
- *      device: {Device.Device},
- *      size: {Number} size in bytes,
- *      input_stream: {Gio.InputStream} readable stream for uploads,
- *      output_stream: {Gio.OutputStream} writable stream for downloads
- *  });
+ * File Transfer base class
  */
-var Transfer = GObject.registerClass({
-    GTypeName: 'GSConnectCoreTransfer',
-    Signals: {
-        'progress': {
-            flags: GObject.SignalFlags.RUN_FIRST,
-            param_types: [ GObject.TYPE_INT ]
-        }
-    },
-    Properties: {
-        'device': GObject.ParamSpec.object(
-            'device',
-            'TransferDevice',
-            'The device associated with this transfer',
-            GObject.ParamFlags.READWRITE | GObject.ParamFlags.CONSTRUCT_ONLY,
-            GObject.Object
-        ),
-        'size': GObject.ParamSpec.uint(
-            'size',
-            'TransferSize',
-            'The size in bytes of the transfer',
-            GObject.ParamFlags.READWRITE | GObject.ParamFlags.CONSTRUCT_ONLY,
-            0, GLib.MAXUINT32,
-            0
-        ),
-        'uuid': GObject.ParamSpec.string(
-            'uuid',
-            'TransferUUID',
-            'The UUID of this transfer',
-            GObject.ParamFlags.READABLE,
-            ''
-        )
-    }
-}, class Transfer extends Channel {
-
-    _init(params) {
-        super._init(params.device.id, 'transfer');
-
-        this._cancellable = new Gio.Cancellable();
-
-        this._device = params.device;
-        this.input_stream = params.input_stream;
-        this.output_stream = params.output_stream;
-        this._size = params.size;
-
-        this.checksum = params.checksum;
-        this._checksum = new GLib.Checksum(GLib.ChecksumType.MD5);
-        this._written = 0;
-        this._progress = 0;
+var Transfer = class Transfer extends Channel {
 
+    /**
+     * @param {object} params - Transfer parameters
+     * @param {Device.Device} params.device - The device that owns this transfer
+     * @param {Gio.InputStream} params.input_stream - The input stream (read)
+     * @param {Gio.OutputStrea} params.output_stream - The output stream (write)
+     * @param {number} params.size - The size of the transfer in bytes
+     */
+    constructor(params) {
+        super(params);
         this.device._transfers.set(this.uuid, this);
     }
 
-    get device() {
-        return this._device;
+    get identity() {
+        return this.device._channel.identity;
     }
 
-    get size() {
-        return this._size || 0;
+    get type() {
+        return 'transfer';
     }
 
+    // For bluetooth transfers this also serves as the per-transfer profile UUID
     get uuid() {
         if (this._uuid === undefined) {
             this._uuid = GLib.uuid_string_random();
@@ -555,6 +498,10 @@ var Transfer = GObject.registerClass({
         return this._uuid;
     }
 
+    set uuid(uuid) {
+        this._uuid = uuid;
+    }
+
     /**
      * Override in protocol implementation
      */
@@ -562,9 +509,6 @@ var Transfer = GObject.registerClass({
         throw new GObject.NotImplementedError();
     }
 
-    /**
-     * Override in protocol implementation
-     */
     async download() {
         throw new GObject.NotImplementedError();
     }
@@ -573,32 +517,16 @@ var Transfer = GObject.registerClass({
      * Cancel the transfer in progress
      */
     cancel() {
-        this._cancellable.cancel();
+        this.close();
     }
 
     close() {
-        super.close();
         this.device._transfers.delete(this.uuid);
+        super.close();
     }
 
     /**
-     * Emit the progress signal with an integer between 0-100
-     * @param {Number} increment - The increment to emit progress signals at
-     */
-    progress(increment=1) {
-        let progress = Math.floor(this._written / this.size * 100);
-
-        if (progress - this._progress >= increment) {
-            this.emit('progress', progress);
-        }
-
-        this._progress = progress;
-    }
-
-    /**
-     * For transfers without a checksum (currently all but notification icons)
-     * we use g_output_stream_splice() which basically loops our read/writes
-     * in 8192 byte chunks.
+     * Transfer using g_output_stream_splice()
      *
      * @return {Boolean} - %true on success, %false on failure.
      */
@@ -611,14 +539,14 @@ var Transfer = GObject.registerClass({
                     this.input_stream,
                     Gio.OutputStreamSpliceFlags.NONE,
                     GLib.PRIORITY_DEFAULT,
-                    this._cancellable,
+                    this.cancellable,
                     (source, res) => {
                         try {
                             if (source.splice_finish(res) < this.size) {
-                                throw new Error('expected more');
-                            } else {
-                                resolve(true);
+                                throw new Error('incomplete data');
                             }
+
+                            resolve(true);
                         } catch (e) {
                             reject(e);
                         }
@@ -633,5 +561,5 @@ var Transfer = GObject.registerClass({
             return result;
         }
     }
-});
+}
 
