@@ -24,14 +24,13 @@ const GLib = imports.gi.GLib;
 const GObject = imports.gi.GObject;
 const Gtk = imports.gi.Gtk;
 
-// Bootstrap
-function getPath() {
-    // Diced from: https://github.com/optimisme/gjs-examples/
-    let m = new RegExp('@(.+):\\d+').exec((new Error()).stack.split('\n')[1]);
+// Find the root datadir of the extension
+function get_datadir() {
+    let m = /@(.+):\d+/.exec((new Error()).stack.split('\n')[1]);
     return Gio.File.new_for_path(m[1]).get_parent().get_parent().get_path();
 }
 
-window.gsconnect = { extdatadir: getPath() };
+window.gsconnect = { extdatadir: get_datadir() };
 imports.searchPath.unshift(gsconnect.extdatadir);
 imports._gsconnect;
 
@@ -40,20 +39,19 @@ const Bluetooth = imports.service.bluetooth;
 const Core = imports.service.core;
 const Device = imports.service.device;
 const Lan = imports.service.lan;
-const Notification = imports.modules.notification;
-const Sms = imports.modules.sms;
-const Sound = imports.modules.sound;
 
 const ServiceUI = imports.service.ui.service;
 const Settings = imports.service.ui.settings;
 
+const _GITHUB = 'https://github.com/andyholmes/gnome-shell-extension-gsconnect';
 
-var Daemon = GObject.registerClass({
-    GTypeName: 'GSConnectDaemon',
+
+const Service = GObject.registerClass({
+    GTypeName: 'GSConnectService',
     Properties: {
         'devices': GObject.param_spec_variant(
             'devices',
-            'DevicesList',
+            'Devices',
             'A list of known devices',
             new GLib.VariantType('as'),
             null,
@@ -68,20 +66,20 @@ var Daemon = GObject.registerClass({
         ),
         'name': GObject.ParamSpec.string(
             'name',
-            'DeviceName',
+            'deviceName',
             'The name announced to the network',
             GObject.ParamFlags.READWRITE,
             'GSConnect'
         ),
         'type': GObject.ParamSpec.string(
             'type',
-            'DeviceType',
+            'deviceType',
             'The service device type',
             GObject.ParamFlags.READABLE,
             'desktop'
         )
     }
-}, class Daemon extends Gtk.Application {
+}, class Service extends Gtk.Application {
 
     _init() {
         super._init({
@@ -101,32 +99,14 @@ var Daemon = GObject.registerClass({
 
     // Properties
     get certificate() {
-        // https://github.com/KDE/kdeconnect-kde/blob/master/core/kdeconnectconfig.cpp#L119
         if (this._certificate === undefined) {
             let certPath = gsconnect.configdir + '/certificate.pem';
-            let certExists = GLib.file_test(certPath, GLib.FileTest.EXISTS);
             let keyPath = gsconnect.configdir + '/private.pem';
-            let keyExists = GLib.file_test(keyPath, GLib.FileTest.EXISTS);
 
-            if (!keyExists || !certExists) {
-                let proc = new Gio.Subprocess({
-                    argv: [
-                        'openssl', 'req',
-                        '-new', '-x509', '-sha256',
-                        '-out', certPath,
-                        '-newkey', 'rsa:2048', '-nodes',
-                        '-keyout', keyPath,
-                        '-days', '3650',
-                        '-subj', '/O=andyholmes.github.io/OU=GSConnect/CN=' + GLib.uuid_string_random()
-                    ],
-                    flags: Gio.SubprocessFlags.STDOUT_SILENCE | Gio.SubprocessFlags.STDERR_SILENCE
-                });
-                proc.init(null);
-                proc.wait_check(null);
-            }
-
-            // Load the certificate
-            this._certificate = Gio.TlsCertificate.new_from_files(certPath, keyPath);
+            this._certificate = Gio.TlsCertificate.new_for_paths(
+                certPath,
+                keyPath
+            );
         }
 
         return this._certificate;
@@ -149,7 +129,7 @@ var Daemon = GObject.registerClass({
                     deviceId: this.certificate.common_name,
                     deviceName: gsconnect.settings.get_string('public-name'),
                     deviceType: this.type,
-                    tcpPort: this.lanService.port,
+                    tcpPort: 1716,
                     protocolVersion: 7,
                     incomingCapabilities: [],
                     outgoingCapabilities: []
@@ -177,11 +157,14 @@ var Daemon = GObject.registerClass({
     get type() {
         if (this._type === undefined) {
             try {
-                let type = Number(
-                    GLib.file_get_contents('/sys/class/dmi/id/chassis_type')[1]
-                );
+                let type = GLib.file_get_contents('/sys/class/dmi/id/chassis_type')[1];
 
-                this._type = ([8, 9, 10, 14].includes(type)) ? 'laptop' : 'desktop';
+                if (type instanceof Uint8Array) {
+                    type = imports.byteArray.toString(type);
+                }
+
+                type = Number(type);
+                this._type = [8, 9, 10, 14].includes(type) ? 'laptop' : 'desktop';
             } catch (e) {
                 this._type = 'desktop';
             }
@@ -191,21 +174,27 @@ var Daemon = GObject.registerClass({
     }
 
     /**
-     * Discovery Methods
+     * Send identity to @address or broadcast if %null
+     *
+     * @param {string|Gio.InetSocketAddress} - TCP address, bluez path or %null
      */
     broadcast(address=null) {
-        switch (true) {
-            case (address instanceof Gio.InetSocketAddress):
-                this.lanService.broadcast(address);
-                break;
+        try {
+            switch (true) {
+                case (address instanceof Gio.InetSocketAddress):
+                    this.lan.broadcast(address);
+                    break;
 
-            case (typeof address === 'string'):
-                this.bluetoothService.broadcast(address);
-                break;
+                case (typeof address === 'string'):
+                    this.bluetooth.broadcast(address);
+                    break;
 
-            default:
-                this.lanService.broadcast();
-                this.bluetoothService.broadcast();
+                // We only do true "broadcasts" for LAN
+                default:
+                    this.lan.broadcast();
+            }
+        } catch (e) {
+            logError(e);
         }
     }
 
@@ -220,66 +209,55 @@ var Daemon = GObject.registerClass({
         let device = this._devices.get(packet.body.deviceId);
 
         if (device === undefined) {
-            log(`GSConnect: Adding ${packet.body.deviceName}`);
+            debug(`GSConnect: Adding ${packet.body.deviceName}`);
+
+            // TODO: Remove when all clients support bluetooth-like discovery
+            //
+            // If this is the third device to connect, disable discovery to
+            // avoid choking on networks with a large amount of devices
+            if (this._devices.size === 2 && this.discoverable) {
+                this.activate_action('discoverable', null);
+
+                let error = new Error();
+                error.name = 'DiscoveryWarning';
+                this.notify_error(error);
+            }
 
             device = new Device.Device(packet);
-
-            device._pruneId = device.connect(
-                'notify::connected',
-                this._pruneDevices.bind(this)
-            );
-
             this._devices.set(device.id, device);
+
+            gsconnect.settings.set_strv('devices', this.devices);
             this.notify('devices');
 
-            let cached = gsconnect.settings.get_strv('devices');
-
-            if (!cached.includes(device.id)) {
-                cached.push(device.id);
-                gsconnect.settings.set_strv('devices', cached);
-            }
+            device.loadPlugins();
         }
 
         return device;
     }
 
-    async _removeDevice(id) {
-        let device = this._devices.get(id);
-
-        if (device) {
-            log(`GSConnect: Removing ${device.name}`);
-
-            device.destroy();
-            this._devices.delete(id);
-            this.notify('devices');
-        }
-
-        return id;
-    }
-
     async _pruneDevices() {
-        // Don't prune devices while the settings window is open
+        // Don't prune devices while the settings window is open; this also
+        // prevents devices from being pruned while being deleted.
         if (this._window && this._window.visible) {
             return;
         }
 
-        let cached = gsconnect.settings.get_strv('devices');
-
-        for (let device of this._devices.values()) {
-            if (!device.connected && !device.paired && cached.includes(device.id)) {
-                device.disconnect(device._pruneId);
-                let id = await this._removeDevice(device.id);
-                cached.splice(cached.indexOf(id), 1);
-                gsconnect.settings.set_strv('devices', cached);
+        for (let [id, device] of this._devices.entries()) {
+            if (!device.connected && !device.paired) {
+                device.destroy();
+                this._devices.delete(id);
+                gsconnect.settings.set_strv('devices', this.devices);
             }
         }
+
+        this.notify('devices');
     }
 
     /**
      * Delete a known device.
      *
-     * This will remove the device from the cache of known devices, then unpair,
-     * disconnect and delete all GSettings and cached files.
+     * Removes the device from the list of known devices, unpairs it, destroys
+     * it and deletes all GSettings and cached files.
      *
      * @param {String} id - The id of the device to delete
      */
@@ -287,31 +265,126 @@ var Daemon = GObject.registerClass({
         let device = this._devices.get(id);
 
         if (device) {
-            let cached = gsconnect.settings.get_strv('devices');
-
-            if (cached.includes(id)) {
-                cached.splice(cached.indexOf(id), 1);
-                gsconnect.settings.set_strv('devices', cached);
-            }
-
             // Stash the settings path before unpairing and removing
             let settings_path = device.settings.path;
-            device.unpair();
-            this._removeDevice(id);
+            device.sendPacket({ type: 'kdeconnect.pair', pair: 'false' });
+
+            //
+            device.destroy();
+            this._devices.delete(id);
 
             // Delete all GSettings
             GLib.spawn_command_line_async(`dconf reset -f ${settings_path}`);
 
             // Delete the cache
             let cache = GLib.build_filenamev([gsconnect.cachedir, id]);
-            GLib.spawn_command_line_async(`rm -rf ${cache}`);
+            Gio.File.rm_rf(cache);
+
+            // Notify
+            gsconnect.settings.set_strv('devices', this.devices);
+            this.notify('devices');
         }
     }
 
-    _connectDeviceAction() {
+    /**
+     * Service GActions
+     */
+    _initActions() {
+        let actions = [
+            // Device
+            ['deviceAction', this._deviceAction.bind(this), '(ssbv)'],
+
+            // App Menu
+            ['connect', this._connectAction.bind(this)],
+            ['preference', this._preferencesAction.bind(this), 's'],
+            ['preferences', this._preferencesAction.bind(this)],
+            ['about', this._aboutAction.bind(this)],
+
+            // Misc service actions
+            ['broadcast', this.broadcast.bind(this)],
+            ['error', this._errorAction.bind(this), 'a{ss}'],
+            ['log', this._logAction.bind(this)],
+            ['debugger', this._debuggerAction.bind(this)],
+            ['wiki', this._wikiAction.bind(this), 's'],
+            ['quit', () => this.quit()]
+        ];
+
+        for (let [name, callback, type] of actions) {
+            let action = new Gio.SimpleAction({
+                name: name,
+                parameter_type: (type) ? new GLib.VariantType(type) : null
+            });
+            action.connect('activate', callback);
+            this.add_action(action);
+        }
+
+        this.add_action(gsconnect.settings.create_action('discoverable'));
+    }
+
+    /**
+     * A wrapper for Device GActions. This is used to route device notification
+     * actions to their device, since GNotifications need an 'app' level action.
+     *
+     * @param {Gio.Action} action - ...
+     * @param {GLib.Variant(av)} parameter - ...
+     * @param {GLib.Variant(s)} parameter[0] - Device Id or '*' for all
+     * @param {GLib.Variant(s)} parameter[1] - GAction name
+     * @param {GLib.Variant(b)} parameter[2] - %false if the parameter is null
+     * @param {GLib.Variant(v)} parameter[3] - GAction parameter
+     */
+    _deviceAction(action, parameter) {
+        parameter = parameter.unpack();
+
+        let id = parameter[0].unpack();
+        let devices = (id === '*') ? this._devices.values() : [this._devices.get(id)];
+
+        for (let device of devices) {
+            // If the device is available
+            if (device) {
+                device.activate_action(
+                    parameter[1].unpack(),
+                    parameter[2].unpack() ? parameter[3].unpack() : null
+                );
+            }
+        }
+    }
+
+    _connectAction() {
         (new ServiceUI.DeviceConnectDialog()).show_all();
     }
 
+    _preferencesAction(page=null, parameter=null) {
+        if (parameter instanceof GLib.Variant) {
+            page = parameter.unpack();
+        }
+
+        if (!this._window) {
+            this._window = new Settings.Window({ application: this });
+            this._window.connect('delete-event', (window) => {
+                window.visible = false;
+                this._pruneDevices();
+                System.gc();
+
+                return true;
+            });
+        }
+
+        // Open to a specific page
+        if (page) {
+            this._window.switcher.foreach(row => {
+                if (row.get_name() === page) {
+                    this._window.switcher.select_row(row);
+                    return;
+                }
+            });
+        // Open the main page
+        } else {
+            this._window._onPrevious();
+        }
+
+        this._window.present();
+    }
+
     _aboutAction() {
         if (this._about === undefined) {
             this._about = new Gtk.AboutDialog({
@@ -335,112 +408,76 @@ var Daemon = GObject.registerClass({
                 website: gsconnect.metadata.url,
                 license_type: Gtk.License.GPL_2_0
             });
-            this._about.connect('delete-event', dialog => dialog.hide_on_delete());
+            this._about.connect('delete-event', () => this._about.hide_on_delete());
         }
 
         this._about.modal = (this._window && this._window.visible);
-        this._about.transient_for = modal ? this._window : null;
+        this._about.transient_for = this._about.modal ? this._window : null;
         this._about.present();
     }
 
-    /**
-     * A meta-action for routing device actions.
-     *
-     * @param {Gio.Action} action - ...
-     * @param {GLib.Variant(av)} parameter - ...
-     * @param {GLib.Variant(s)} parameter[0] - Device Id
-     * @param {GLib.Variant(s)} parameter[1] - GAction name
-     * @param {GLib.Variant(b)} parameter[2] - %false if the parameter is null
-     * @param {GLib.Variant(v)} parameter[3] - GAction parameter
-     */
-    _deviceAction(action, parameter) {
-        parameter = parameter.unpack();
+    _errorAction(action, parameter) {
+        try {
+            let error = parameter.deep_unpack();
+            let dialog = new Gtk.MessageDialog({
+                text: error.message.trim(),
+                secondary_text: error.stack.trim(),
+                buttons: Gtk.ButtonsType.CLOSE,
+                message_type: Gtk.MessageType.ERROR,
+            });
+            let issues = dialog.add_button(_('Report'), 1);
+            dialog.set_keep_above(true);
 
-        let id = parameter[0].unpack();
-        let devices = (id === '*') ? this._devices.values() : [this._devices.get(id)];
+            let [message, stack] = dialog.get_message_area().get_children();
+            message.halign = Gtk.Align.START;
+            message.selectable = true;
+            stack.selectable = true;
 
-        for (let device of devices) {
-            // If the device is available
-            if (device) {
-                device.activate_action(
-                    parameter[1].unpack(),
-                    parameter[2].unpack() ? parameter[3].unpack() : null
-                );
-            }
+            dialog.connect('response', (dialog, id) => {
+                if (id === 1) {
+                    let query = encodeURIComponent(dialog.text).replace('%20', '+');
+                    this._github(`issues?q=is%3Aissue+${query}`);
+                } else {
+                    dialog.destroy();
+                }
+            });
+
+            dialog.show();
+        } catch (e) {
+            logError(e);
         }
     }
 
-    /**
-     * Service GActions
-     */
-    _initActions() {
-        let actions = [
-            // Device
-            ['deviceAction', this._deviceAction.bind(this), '(ssbv)'],
-            // Daemon
-            ['connectDevice', this._connectDeviceAction.bind(this)],
-            ['broadcast', this.broadcast.bind(this)],
-            ['openSettings', this.openSettings.bind(this)],
-            ['quit', this.quit.bind(this)],
-            ['debugger', this._debugger.bind(this)],
-            ['about', this._aboutAction.bind(this)]
-        ];
+    _logAction() {
+        // Ensure debugging is enabled
+        gsconnect.settings.set_boolean('debug', true);
 
-        actions.map(entry => {
-            let action = new Gio.SimpleAction({
-                name: entry[0],
-                parameter_type: (entry[2]) ? new GLib.VariantType(entry[2]) : null
-            });
-            action.connect('activate', entry[1]);
-            this.add_action(action);
-        });
-
-        this.add_action(
-            new Gio.PropertyAction({
-                name: 'discoverable',
-                object: this,
-                property_name: 'discoverable'
-            })
+        // Launch a terminal with tabs for GJS and Gnome Shell
+        GLib.spawn_command_line_async(
+            'gnome-terminal ' +
+            `--tab --title "GJS" --command "journalctl -f -o cat /usr/bin/gjs" ` +
+            '--tab --title "Gnome Shell" --command "journalctl -f -o cat /usr/bin/gnome-shell"'
         );
     }
 
-    _debugger() {
-        (new imports.modules.debug.Window()).present();
+    _debuggerAction() {
+        (new imports.service.components.debug.Window()).present();
     }
 
-    /**
-     * Open the application settings, for @device if given.
-     * @param {String} [device] - Device id
-     *
-     * The DBus method takes no arguments; Device.openSettings() calls this and
-     * populates @device.
-     */
-    openSettings(device=null) {
-        if (!this._window) {
-            this._window = new Settings.Window({ application: this });
-            this._window.connect('delete-event', (window) => {
-                window.visible = false;
-                this._pruneDevices();
-                System.gc();
+    _wikiAction(action, parameter) {
+        this._github(`wiki/${parameter.unpack()}`);
+    }
 
-                return true;
-            });
-        }
+    _github(path=[]) {
+        let uri = [_GITHUB].concat(path.split('/')).join('/');
 
-        this._window.present();
-
-        // Open to a device page
-        if (device) {
-            this._window.switcher.foreach(row => {
-                if (row.get_name() === device) {
-                    this._window.switcher.select_row(row);
-                    return;
-                }
-            });
-        // Open the main page
-        } else {
-            this._window._onPrevious();
-        }
+        Gio.AppInfo.launch_default_for_uri_async(uri, null, null, (src, res) => {
+            try {
+                Gio.AppInfo.launch_default_for_uri_finish(res);
+            } catch (e) {
+                logError(e);
+            }
+        });
     }
 
     /**
@@ -470,7 +507,7 @@ var Daemon = GObject.registerClass({
      * @param {String|Number} id - Gtk (string) or libnotify id (uint32)
      * @param {String|null} application - Application Id if Gtk or null
      */
-    async remove_notification(id, application=null) {
+    remove_notification(id, application=null) {
         let name, path, method, variant;
 
         if (application !== null) {
@@ -496,8 +533,161 @@ var Daemon = GObject.registerClass({
         });
     }
 
+    /**
+     * Report a service-level error
+     *
+     * @param {object} error - An Error or object with name, message and stack
+     * @param {string} context - The scope of the error
+     */
+    notify_error(error) {
+        try {
+            // Always log the error
+            logError(error);
+
+            // Create an new notification
+            let id, title, body, icon, action;
+            let notif = new Gio.Notification();
+            notif.set_priority(Gio.NotificationPriority.URGENT);
+
+            switch (error.name) {
+                // A TLS certificate failure
+                case 'AuthenticationError':
+                    id = `${Date.now()}`;
+                    title = _('Authentication Failure');
+                    let time = GLib.DateTime.new_now_local().format('%F %R');
+                    body = `"${error.deviceName}"@${error.deviceHost} (${time})`;
+                    icon = new Gio.ThemedIcon({ name: 'dialog-error' });
+                    break;
+
+                case 'LanError':
+                case 'ProxyError':
+                    id = error.name;
+                    title = _('Network Error');
+                    body = error.message + '\n\n' + _('Click for help troubleshooting');
+                    icon = new Gio.ThemedIcon({ name: 'network-error' });
+                    notif.set_default_action(
+                        `app.wiki('Troubleshooting#${error.name}')`
+                    );
+                    break;
+
+                case 'GvcError':
+                    id = error.name;
+                    title = _('PulseAudio Error');
+                    body = _('Click for help troubleshooting');
+                    icon = new Gio.ThemedIcon({ name: 'dialog-error' });
+                    notif.set_default_action(
+                        `app.wiki('Troubleshooting#${error.name}')`
+                    );
+                    break;
+
+                case 'DiscoveryWarning':
+                    id = 'discovery-warning';
+                    title = _('Discovery Disabled');
+                    body = _('Discovery has been disabled due to the number of devices on this network.') +
+                           '\n\n' +
+                           _('Click to open preferences');
+                    icon = new Gio.ThemedIcon({ name: 'dialog-warning' });
+                    notif.set_default_action('app.preference::service');
+                    notif.set_priority(Gio.NotificationPriority.NORMAL);
+                    break;
+
+                // Missing sshfs, libcanberra
+                case 'DependencyError':
+                    id = 'dependency-error';
+                    title = _('Additional Software Required');
+                    body = _('Click to open preferences');
+                    icon = new Gio.ThemedIcon({ name: 'system-software-install-symbolic' });
+                    notif.set_default_action('app.preference::other');
+                    notif.set_priority(Gio.NotificationPriority.HIGH);
+                    break;
+
+                case 'PluginError':
+                    id = `${error.plugin}-error`;
+                    title = _('%s Plugin Failed To Load').format(error.label);
+                    body = error.message + '\n\n' + _('Click for more information');
+                    icon = new Gio.ThemedIcon({ name: 'dialog-error' });
+
+                    error = new GLib.Variant('a{ss}', {
+                        name: error.name,
+                        message: error.message,
+                        stack: error.stack
+                    });
+
+                    notif.set_default_action_and_target('app.error', error);
+                    notif.set_priority(Gio.NotificationPriority.HIGH);
+                    break;
+
+                case 'SSHSignatureError':
+                    id = error.name;
+                    title = _('Remote Filesystem Error');
+                    body = _('%s is using an incompatible SSH library').format(error.deviceName) + '\n\n' +
+                           _('Click for more information');
+                    icon = new Gio.ThemedIcon({ name: 'dialog-error' });
+                    notif.set_default_action(
+                        `app.wiki('Troubleshooting#${error.name}')`
+                    );
+                    notif.set_priority(Gio.NotificationPriority.HIGH);
+                    break;
+
+                case 'WaylandNotSupported':
+                    id = error.name;
+                    title = _('Wayland Not Supported');
+                    body = _('Remote input not supported on Wayland') + '\n\n' +
+                           _('Click for more information');
+                    icon = new Gio.ThemedIcon({ name: 'preferences-desktop-display-symbolic' });
+                    notif.set_default_action(
+                        `app.wiki('Troubleshooting#${error.name}')`
+                    );
+                    notif.set_priority(Gio.NotificationPriority.HIGH);
+                    break;
+
+                default:
+                    id = `${Date.now()}`;
+                    title = error.name;
+                    body = error.message.trim();
+                    icon = new Gio.ThemedIcon({ name: 'dialog-error' });
+                    error = new GLib.Variant('a{ss}', {
+                        name: error.name,
+                        message: error.message,
+                        stack: error.stack
+                    });
+                    notif.set_default_action_and_target('app.error', error);
+                    notif.set_priority(Gio.NotificationPriority.HIGH);
+            }
+
+            // Create an urgent notification
+            notif.set_title(_('GSConnect: %s').format(title));
+            notif.set_body(body);
+            notif.set_icon(icon);
+
+            // Bypass override
+            super.send_notification(id, notif);
+        } catch (e) {
+            logError(e);
+        }
+    }
+
+    /**
+     * Load each script in components/ and instantiate a Service if it has one
+     */
+    _loadComponents() {
+        for (let name in imports.service.components) {
+            try {
+                let module = imports.service.components[name];
+
+                if (module.hasOwnProperty('Service')) {
+                    this[name] = new module.Service();
+                }
+            } catch (e) {
+                this.notify_error(e);
+            }
+        }
+    }
+
     vfunc_activate() {
-        this.broadcast();
+        // TODO: this causes problems right now because the bluetooth service
+        // clobbers open TCP channels sometimes, and this gets called often
+        //this.broadcast();
     }
 
     vfunc_startup() {
@@ -505,14 +695,23 @@ var Daemon = GObject.registerClass({
 
         this.hold();
 
-        // Watch *this* file and stop the daemon if it's changed/removed
-        this._daemonMonitor = Gio.File.new_for_path(
+        // Watch *this* file and stop the service if it's updated/uninstalled
+        this._serviceMonitor = Gio.File.new_for_path(
             gsconnect.extdatadir + '/service/daemon.js'
         ).monitor(
             Gio.FileMonitorFlags.WATCH_MOVES,
             null
         );
-        this._daemonMonitor.connect('changed', () => this.quit());
+        this._serviceMonitor.connect('changed', () => this.quit());
+
+        // Init some resources
+        let provider = new Gtk.CssProvider();
+        provider.load_from_resource(gsconnect.app_path + '/application.css');
+        Gtk.StyleContext.add_provider_for_screen(
+            Gdk.Screen.get_default(),
+            provider,
+            Gtk.STYLE_PROVIDER_PRIORITY_APPLICATION
+        );
 
         // Properties
         gsconnect.settings.bind(
@@ -529,47 +728,46 @@ var Daemon = GObject.registerClass({
             Gio.SettingsBindFlags.DEFAULT
         );
 
-        // Init some resources
-        let provider = new Gtk.CssProvider();
-        provider.load_from_resource(gsconnect.app_path + '/application.css');
-        Gtk.StyleContext.add_provider_for_screen(
-            Gdk.Screen.get_default(),
-            provider,
-            Gtk.STYLE_PROVIDER_PRIORITY_APPLICATION
-        );
+        // Keep identity updated and broadcast any name changes
+        gsconnect.settings.connect('changed::public-name', (settings) => {
+            this.identity.body.deviceName = this.name;
+        });
+
+        // GActions
+        this._initActions();
+
+        // Components (PulseAudio, UPower, etc)
+        this._loadComponents();
 
         // Track devices with id as key
         this._devices = new Map();
 
         // Load cached devices
         let cached = gsconnect.settings.get_strv('devices');
-        log(`Loading ${cached.length} device(s) from cache`);
-        cached.map(id => this._ensureDevice({
-            body: {
-                deviceId: id,
-                deviceName: 'cached device'
-            }
-        }));
+        debug(`Loading ${cached.length} device(s) from cache`);
+        cached.map(id => {
+            let device = new Device.Device({ body: { deviceId: id } });
+            this._devices.set(device.id, device);
+            device.loadPlugins();
+        });
+        this.notify('devices');
 
         // Lan.ChannelService
         try {
-            this.lanService = new Lan.ChannelService();
+            this.lan = new Lan.ChannelService();
         } catch (e) {
-            logError(e, 'Lan.ChannelService');
+            e.name = 'LanError';
+            this.notify_error(e);
         }
 
         // Bluetooth.ChannelService
         try {
-            this.bluetoothService = new Bluetooth.ChannelService();
+            this.bluetooth = new Bluetooth.ChannelService();
         } catch (e) {
-            logError(e, 'Bluetooth.ChannelService');
+            if (this.bluetooth) {
+                this.bluetooth.destroy();
+            }
         }
-
-        // GActions
-        this._initActions();
-
-        // Notification Listener
-        this.notificationListener = new Notification.Listener();
     }
 
     vfunc_dbus_register(connection, object_path) {
@@ -595,7 +793,6 @@ var Daemon = GObject.registerClass({
         super.vfunc_dbus_unregister(connection, object_path);
     }
 
-    // FIXME: this is all garbage
     vfunc_open(files, hint) {
         super.vfunc_open(files, hint);
 
@@ -605,16 +802,15 @@ var Daemon = GObject.registerClass({
 
             try {
                 if (file.get_uri_scheme() === 'sms') {
-                    //let uri = new Sms.URI(file.get_uri()); // TODO
                     title = _('Send SMS');
                     action = 'uriSms';
                     parameter = new GLib.Variant('s', file.get_uri());
                 } else if (file.get_uri_scheme() === 'tel') {
                     title = _('Dial Number');
-                    action = 'shareUrl';
+                    action = 'shareUri';
                     parameter = new GLib.Variant('s', file.get_uri());
                 } else {
-                    throw new Error('Unsupported file type');
+                    throw new Error('Unsupported file/URI type');
                 }
 
                 for (let device of this._devices.values()) {
@@ -638,7 +834,7 @@ var Daemon = GObject.registerClass({
                     win.destroy();
                 }
             } catch (e) {
-                logError(e, 'Opening file');
+                logError(e, `GSConnect: Opening ${file.get_uri()}:`);
             }
         }
     }
@@ -646,13 +842,28 @@ var Daemon = GObject.registerClass({
     vfunc_shutdown() {
         super.vfunc_shutdown();
 
-        log('GSConnect: Shutting down');
+        this._devices.forEach(device => device.destroy());
 
-        this.lanService.destroy();
-        this.bluetoothService.destroy();
-        this.notificationListener.destroy();
+        if (this.mpris) {
+            this.mpris.destroy();
+        }
+
+        if (this.notification) {
+            this.notification.destroy();
+        }
+
+        if (this.lan) {
+            this.lan.destroy();
+        }
+
+        // FIXME: Really, really bad hack, but sometimes hangs in bluez can
+        // prevent the service from stopping or even hang the desktop.
+        System.exit(0);
+        if (this.bluetooth) {
+            this.bluetooth.destroy();
+        }
     }
 });
 
-(new Daemon()).run([System.programInvocationName].concat(ARGV));
+(new Service()).run([System.programInvocationName].concat(ARGV));
 
