@@ -6,36 +6,14 @@ const GObject = imports.gi.GObject;
 const Gtk = imports.gi.Gtk;
 
 const Core = imports.service.core;
-const DBus = imports.modules.dbus;
+const DBus = imports.service.components.dbus;
+const Lan = imports.service.lan;
 const Bluetooth = imports.service.bluetooth;
 
 const UUID = 'org.gnome.Shell.Extensions.GSConnect.Device';
 const INTERFACE_INFO = gsconnect.dbusinfo.lookup_interface(UUID);
 
 
-/**
- * Base class for plugin actions
- */
-var Action = GObject.registerClass({
-    GTypeName: 'GSConnectDeviceAction'
-}, class Action extends Gio.SimpleAction {
-
-    _init(params) {
-        super._init({
-            name: params.name,
-            parameter_type: params.parameter_type
-        });
-
-        this.summary = params.summary;
-        this.description = params.description;
-        this.icon_name = params.icon_name;
-        this.incoming = params.incoming;
-        this.outgoing = params.outgoing;
-        this.allow = params.allow;
-    }
-});
-
-
 /**
  * An object representing a remote device.
  *
@@ -49,37 +27,29 @@ var Device = GObject.registerClass({
     Properties: {
         'connected': GObject.ParamSpec.boolean(
             'connected',
-            'deviceConnected',
+            'Connected',
             'Whether the device is connected',
             GObject.ParamFlags.READABLE,
-            false
+            null
         ),
         'encryption-info': GObject.ParamSpec.string(
             'encryption-info',
             'Encryption Info',
             'A formatted string with the local and remote fingerprints',
             GObject.ParamFlags.READABLE,
-            ''
-        ),
-        'errors': GObject.param_spec_variant(
-            'errors',
-            'Device Errors',
-            'A list of device errors',
-            new GLib.VariantType('as'),
-            null,
-            GObject.ParamFlags.READABLE
+            null
         ),
         'icon-name': GObject.ParamSpec.string(
             'icon-name',
-            'IconName',
-            'Icon name representing the service device',
+            'Icon Name',
+            'Icon name representing the device',
             GObject.ParamFlags.READABLE,
-            ''
+            null
         ),
         'id': GObject.ParamSpec.string(
             'id',
             'deviceId',
-            'The device id/hostname',
+            'The device hostname or other unique id',
             GObject.ParamFlags.READABLE,
             ''
         ),
@@ -88,28 +58,21 @@ var Device = GObject.registerClass({
             'deviceName',
             'The device name',
             GObject.ParamFlags.READABLE,
-            ''
+            null
         ),
         'paired': GObject.ParamSpec.boolean(
             'paired',
-            'devicePaired',
+            'Paired',
             'Whether the device is paired',
             GObject.ParamFlags.READABLE,
-            false
-        ),
-        'symbolic-icon-name': GObject.ParamSpec.string(
-            'symbolic-icon-name',
-            'SybolicIconName',
-            'Symbolic icon name representing the device',
-            GObject.ParamFlags.READABLE,
-            ''
+            null
         ),
         'type': GObject.ParamSpec.string(
             'type',
             'deviceType',
             'The device type',
             GObject.ParamFlags.READABLE,
-            'unknown'
+            null
         )
     }
 }, class Device extends Gio.SimpleActionGroup {
@@ -117,7 +80,6 @@ var Device = GObject.registerClass({
     _init(identity) {
         super._init();
 
-        this.service = Gio.Application.get_default();
         this._channel = null;
         this._connected = false;
 
@@ -125,10 +87,9 @@ var Device = GObject.registerClass({
         this._incomingPairRequest = 0;
         this._outgoingPairRequest = 0;
 
-        // Maps of name->plugin, packet->plugin, name->Error
+        // Maps of name->Plugin, packet->Plugin, uuid->Transfer
         this._plugins = new Map();
         this._handlers = new Map();
-        this._errors = new Map();
         this._transfers = new Map();
 
         // We at least need the device Id for GSettings and the DBus interface
@@ -140,7 +101,7 @@ var Device = GObject.registerClass({
             path: `/org/gnome/shell/extensions/gsconnect/device/${deviceId}/`
         });
 
-        // TODO: Backwards compatibility; remove after a few releases
+        // TODO: Backwards compatibility <= v12; remove after a few releases
         if (this.settings.get_string('certificate-pem') !== '') {
             this.settings.set_boolean('paired', true);
         }
@@ -164,6 +125,7 @@ var Device = GObject.registerClass({
         this._dbus_object.add_interface(this._dbus);
 
         // GActions/GMenu
+        this._registerActions();
         this._actionsId = Gio.DBus.session.export_action_group(
             this.object_path,
             this
@@ -174,9 +136,6 @@ var Device = GObject.registerClass({
             this.object_path,
             this.menu
         );
-
-        this._registerActions();
-        this._loadPlugins();
     }
 
     /** Device Properties */
@@ -209,25 +168,16 @@ var Device = GObject.registerClass({
             ).fingerprint();
         }
 
-        // TRANSLATORS: Remote and local TLS Certificate fingerprint
-        // PLEASE KEEP NEWLINE CHARACTERS (\n)
+        // TRANSLATORS: Label for TLS Certificate fingerprint
         //
         // Example:
         //
         // Google Pixel Fingerprint:
         // 00:00:00:00:00:00:00:00:00:00:00:00:00:00:00:00:00:00:00:00
-        //
-        // Local Fingerprint:
-        // 00:00:00:00:00:00:00:00:00:00:00:00:00:00:00:00:00:00:00:00
-        return _('%s Fingerprint:\n%s\n\nLocal Fingerprint:\n%s').format(
-            this.name,
-            fingerprint,
-            this.service.fingerprint
-        );
-    }
-
-    get errors() {
-        return this._errors;
+        return _('%s Fingerprint:').format(this.name) + '\n' +
+            fingerprint + '\n\n' +
+            _('%s Fingerprint:').format(this.service.name) + '\n' +
+            this.service.fingerprint;
     }
 
     get id() {
@@ -247,10 +197,10 @@ var Device = GObject.registerClass({
     }
 
     get allowed_plugins() {
-        let blacklist = this.settings.get_strv('plugin-blacklist');
+        let disabled = this.settings.get_strv('disabled-plugins');
 
         return this.supported_plugins.filter(name => {
-            return !blacklist.includes(name);
+            return !disabled.includes(name);
         });
     }
 
@@ -265,8 +215,8 @@ var Device = GObject.registerClass({
         }
     }
 
-    get symbolic_icon_name() {
-        return `${this.icon_name}-symbolic`;
+    get service() {
+        return Gio.Application.get_default();
     }
 
     get type() {
@@ -318,12 +268,10 @@ var Device = GObject.registerClass({
         let supported = [];
 
         for (let name in imports.service.plugins) {
-            // Skip base.js
-            if (name === 'base')
-                continue;
-
             let meta = imports.service.plugins[name].Metadata;
 
+            if (!meta) continue;
+
             // If we can handle packets it sends...
             if (meta.incomingCapabilities.some(t => packet.body.outgoingCapabilities.includes(t))) {
                 supported.push(name);
@@ -336,6 +284,36 @@ var Device = GObject.registerClass({
         this.settings.set_strv('supported-plugins', supported.sort());
     }
 
+    /**
+     * This is invoked by Core.Channel.attach() which also sets this._channel
+     */
+    _setConnected() {
+        debug(`Connected to ${this.name} (${this.id})`);
+
+        this.settings.set_string('last-connection', this._channel.type);
+
+        this._connected = true;
+        this.notify('connected');
+
+        this._plugins.forEach(async (plugin) => plugin.connected());
+    }
+
+    /**
+     * This is the callback for the Core.Channel's cancellable object
+     */
+    _setDisconnected() {
+        debug(`Disconnected from ${this.name} (${this.id})`);
+
+        this._channel = null;
+        this._connected = false;
+        this.notify('connected');
+
+        this._plugins.forEach(async (plugin) => plugin.disconnected());
+
+        // TODO: not ideal calling back and forth like this
+        this.service._pruneDevices();
+    }
+
     /**
      * Request a connection from the device
      */
@@ -375,27 +353,31 @@ var Device = GObject.registerClass({
      * @param {Core.Packet} packet - The incoming packet object
      */
     receivePacket(packet) {
-        let handler = this._handlers.get(packet.type);
+        try {
+            let handler = this._handlers.get(packet.type);
 
-        switch (true) {
-            // We handle pair requests
-            case (packet.type === 'kdeconnect.pair'):
-                this._handlePair(packet);
-                break;
+            switch (true) {
+                // We handle pair requests
+                case (packet.type === 'kdeconnect.pair'):
+                    this._handlePair(packet);
+                    break;
 
-            // The device thinks we're paired; send an unpair packet and bail
-            case !this.paired:
-                this.unpair();
-                break;
+                // The device must think we're paired; inform it we are not
+                case !this.paired:
+                    this.unpair();
+                    break;
 
-            // This is a supported packet
-            case (handler !== undefined):
-                handler.handlePacket(packet);
-                break;
+                // This is a supported packet
+                case (handler !== undefined):
+                    handler.handlePacket(packet);
+                    break;
 
-            // This is an unsupported packet or disabled plugin
-            default:
-                logWarning(`Unsupported packet type (${packet.type})`, this.name);
+                // This is an unsupported packet or disabled plugin
+                default:
+                    throw new Error(`Unsupported packet type (${packet.type})`);
+            }
+        } catch (e) {
+            logError(e, this.name);
         }
     }
 
@@ -405,135 +387,99 @@ var Device = GObject.registerClass({
      * @param {Gio.Stream} payload - A payload stream // TODO
      */
     sendPacket(packet, payload=null) {
-        if (this.connected && (this.paired || packet.type === 'kdeconnect.pair')) {
-            this._channel.send(packet);
+        try {
+            if (this.connected && (this.paired || packet.type === 'kdeconnect.pair')) {
+                this._channel.send(packet);
+            }
+        } catch (e) {
+            logError(e, this.name);
         }
     }
 
-    /** Channel Callbacks */
-    _onConnected(channel) {
-        log(`Connected to ${this.name} (${this.id})`);
-
-        this.settings.set_string('last-connection', channel.type);
-
-        this._connected = true;
-        this.notify('connected');
-        this.notify('symbolic-icon-name');
-    }
-
-    _onDisconnected(channel) {
-        log(`Disconnected from ${this.name} (${this.id})`);
-
-        GObject.signal_handlers_destroy(channel);
-        this._channel = null;
-
-        this._connected = false;
-        this.notify('connected');
-        this.notify('symbolic-icon-name');
-    }
-
     /**
-     * Stock device actions
+     * Actions
      */
     _registerActions() {
-        let activate = new Action({
+        // Stock device actions
+        let activate = new Gio.SimpleAction({
             name: 'activate',
             parameter_type: null,
-            summary: _('Activate'),
-            description: _('Attempt reconnection'),
-            icon_name: 'view-refresh-symbolic'
+            state: new GLib.Variant('(ss)', [_('Reconnect'), 'view-refresh-symbolic']),
         });
         activate.connect('activate', this.activate.bind(this));
         this.add_action(activate);
 
-        let cancelTransfer = new Action({
-            name: 'cancelTransfer',
-            parameter_type: new GLib.VariantType('s'),
-            summary: _('Cancel Transfer'),
-            description: _('Stop an in progress transfer'),
-            icon_name: 'process-stop-symbolic'
-        });
-        cancelTransfer.connect('activate', this.cancelTransfer.bind(this));
-        this.add_action(cancelTransfer);
-
-        let openSettings = new Action({
+        let openSettings = new Gio.SimpleAction({
             name: 'openSettings',
             parameter_type: null,
-            summary: _('Open Settings'),
-            description: _('Open the settings window for the device'),
-            icon_name: 'preferences-system-symbolic'
+            state: new GLib.Variant('(ss)', [_('Settings'), 'preferences-system-symbolic'])
         });
         openSettings.connect('activate', this.openSettings.bind(this));
         this.add_action(openSettings);
 
-        let acceptPair = new Action({
+        let acceptPair = new Gio.SimpleAction({
             name: 'pair',
             parameter_type: null,
-            summary: _('Pair'),
-            description: _('Send or accept a pair request'),
-            icon_name: 'channel-insecure-symbolic'
+            state: new GLib.Variant('(ss)', [_('Pair'), 'channel-secure-symbolic'])
         });
         acceptPair.connect('activate', this.pair.bind(this));
         this.add_action(acceptPair);
 
-        let rejectPair = new Action({
+        let rejectPair = new Gio.SimpleAction({
             name: 'unpair',
             parameter_type: null,
-            summary: _('Unpair'),
-            description: _('Unpair or reject a pair request'),
-            icon_name: 'channel-insecure-symbolic'
+            state: new GLib.Variant('(ss)', [_('Unpair'), 'channel-insecure-symbolic'])
         });
         rejectPair.connect('activate', this.unpair.bind(this));
         this.add_action(rejectPair);
 
-        let viewFolder = new Action({
-            name: 'viewFolder',
-            parameter_type: new GLib.VariantType('s'),
-            summary: _('View Folder'),
-            description: _('Open a folder for viewing'),
-            icon_name: 'folder-open-symbolic'
+        // Some general utility actions
+        let cancelTransfer = new Gio.SimpleAction({
+            name: 'cancelTransfer',
+            parameter_type: new GLib.VariantType('s')
         });
-        viewFolder.connect('activate', this.viewFolder.bind(this));
-        this.add_action(viewFolder);
+        cancelTransfer.connect('activate', this.cancelTransfer.bind(this));
+        this.add_action(cancelTransfer);
+
+        let openPath = new Gio.SimpleAction({
+            name: 'openPath',
+            parameter_type: new GLib.VariantType('s')
+        });
+        openPath.connect('activate', this.openPath.bind(this));
+        this.add_action(openPath);
     }
 
-    cancelTransfer(action, parameter) {
-        let uuid = parameter.unpack();
-        let transfer = this._transfers.get(uuid);
-
-        if (transfer !== undefined) {
-            this._transfers.delete(uuid);
-            transfer.cancel();
-        }
-    }
-
-    viewFolder(action, parameter) {
-        let path = parameter.get_string()[0];
+    openPath(action, parameter) {
+        let path = parameter.unpack();
         path = path.startsWith('file://') ? path : `file://${path}`;
-        Gio.AppInfo.launch_default_for_uri(path, null);
+
+        Gio.AppInfo.launch_default_for_uri_async(path, null, null, (src, res) => {
+            try {
+                Gio.AppInfo.launch_default_for_uri_finish(res);
+            } catch (e) {
+                logError(e);
+            }
+        });
     }
 
     /**
-     * Device notifications
+     * Hide a notification, device analog for GApplication.withdraw_notification()
+     *
+     * @param {string} id - Id for the notification to withdraw
      */
-    send_notification(id, notification) {
-        this.service.send_notification(`${this.id}|${id}`, notification);
-    }
-
-    withdraw_notification(id) {
-        this.service.withdraw_notification(`${this.id}|${id}`);
-    }
-
     hideNotification(id) {
         this.service.withdraw_notification(`${this.id}|${id}`);
     }
 
+    /**
+     * Show a notification, device analog for GApplication.send_notification()
+     */
     showNotification(params) {
         params = Object.assign({
             id: GLib.DateTime.new_now_local().to_unix(),
             title: this.name,
             body: '',
-            icon: new Gio.ThemedIcon({ name: this.symbolic_icon_name }),
+            icon: new Gio.ThemedIcon({ name: `${this.icon_name}-symbolic` }),
             priority: Gio.NotificationPriority.NORMAL,
             action: null,
             buttons: []
@@ -587,6 +533,36 @@ var Device = GObject.registerClass({
         this.service.send_notification(`${this.id}|${params.id}`, notif);
     }
 
+    cancelTransfer(action, parameter) {
+        let uuid = parameter.unpack();
+        let transfer = this._transfers.get(uuid);
+
+        if (transfer !== undefined) {
+            this._transfers.delete(uuid);
+            transfer.cancel();
+        }
+    }
+
+    createTransfer(params) {
+        params.device = this;
+
+        switch (true) {
+            case (this.connection_type === 'tcp'):
+                return new Lan.Transfer(params);
+
+            case (this.connection_type === 'bluetooth'):
+                //return new Bluetooth.Transfer(params);
+
+            // The default is a mock transfer that always appears to fail
+            default:
+                return {
+                    uuid: 'mock-transfer',
+                    download: () => false,
+                    upload: () => false
+                };
+        }
+    }
+
     /**
      * Pair request handler
      *
@@ -597,26 +573,26 @@ var Device = GObject.registerClass({
         if (packet.body.pair) {
             // The device is accepting our request
             if (this._outgoingPairRequest) {
-                log(`Pair accepted by ${this.name}`);
+                debug(`Pair accepted by ${this.name}`);
 
                 this._setPaired(true);
-                this._loadPlugins();
+                this.loadPlugins();
             // The device thinks we're unpaired
             } else if (this.paired) {
                 this._setPaired(true);
                 this.pair();
-                this._loadPlugins();
+                this.loadPlugins();
             // The device is requesting pairing
             } else {
-                log(`Pair request from ${this.name}`);
+                debug(`Pair request from ${this.name}`);
                 this._notifyPairRequest();
             }
         // Device is requesting unpairing/rejecting our request
         } else {
-            log(`Pair rejected by ${this.name}`);
+            debug(`Pair rejected by ${this.name}`);
 
             this._setPaired(false);
-            this._unloadPlugins();
+            this.unloadPlugins();
         }
     }
 
@@ -660,7 +636,7 @@ var Device = GObject.registerClass({
      */
     _resetPairRequest() {
         if (this._incomingPairRequest) {
-            this.withdraw_notification('pair-request');
+            this.hideNotification('pair-request');
             GLib.source_remove(this._incomingPairRequest);
             this._incomingPairRequest = 0;
         }
@@ -679,8 +655,6 @@ var Device = GObject.registerClass({
     _setPaired(bool) {
         this._resetPairRequest();
 
-        this.settings.set_boolean('paired', bool);
-
         // For TCP connections we store or reset the TLS Certificate
         if (this.connection_type === 'tcp') {
             if (bool) {
@@ -693,8 +667,8 @@ var Device = GObject.registerClass({
             }
         }
 
+        this.settings.set_boolean('paired', bool);
         this.notify('paired');
-        this.notify('symbolic-icon-name');
     }
 
     /**
@@ -708,7 +682,7 @@ var Device = GObject.registerClass({
             // then loop back around to send confirmation...
             this.pair();
             // ...before loading plugins
-            this._loadPlugins();
+            this.loadPlugins();
             return;
         }
 
@@ -729,7 +703,7 @@ var Device = GObject.registerClass({
                 this._setPaired.bind(this, false)
             );
 
-            log(`Pair request sent to ${this.name}`);
+            debug(`Pair request sent to ${this.name}`);
         }
     }
 
@@ -749,7 +723,7 @@ var Device = GObject.registerClass({
         }
 
         this._setPaired(false);
-        this._unloadPlugins();
+        this.unloadPlugins();
     }
 
     /**
@@ -777,14 +751,16 @@ var Device = GObject.registerClass({
         return this._plugins.get(name) || null;
     }
 
-    async loadPlugin(name) {
-        debug(`loading '${name}' plugin`, this.name);
+    loadPlugin(name) {
+        let handler, plugin;
+
+        try {
+            if (this.paired && !this._plugins.has(name)) {
+                debug(`loading '${name}' plugin`, this.name);
 
-        if (this.paired && !this._plugins.has(name)) {
-            try {
                 // Instantiate the handler
-                let handler = imports.service.plugins[name];
-                let plugin = new handler.Plugin(this);
+                handler = imports.service.plugins[name];
+                plugin = new handler.Plugin(this);
 
                 // Register packet handlers
                 for (let packetType of handler.Metadata.incomingCapabilities) {
@@ -795,69 +771,90 @@ var Device = GObject.registerClass({
 
                 // Register plugin
                 this._plugins.set(name, plugin);
-                this.errors.delete(name);
-            } catch (e) {
-                logWarning(`loading ${name}: ${e.message}`, this.name);
-                this.errors.set(name, e);
+
+                // Run the connected()/disconnected() handler
+                this.connected ? plugin.connected() : plugin.disconnected();
             }
+        } catch (e) {
+            e.name = (e.name === 'Error') ? 'PluginError' : e.name;
+            e.device = this.id;
+            e.plugin = name;
+            e.label = handler.Metadata.label;
+
+            this.service.notify_error(e);
         }
-
-        this.notify('errors');
     }
 
-    async _loadPlugins() {
-        this.allowed_plugins.map(name => this.loadPlugin(name));
+    async loadPlugins() {
+        for (let name of this.allowed_plugins) {
+            await this.loadPlugin(name);
+        }
     }
 
-    async unloadPlugin(name) {
-        debug(`unloading '${name}' plugin`, this.name);
+    unloadPlugin(name) {
+        let handler, plugin;
 
         try {
-            // Unregister packet handlers
-            let handler = imports.service.plugins[name];
+            if (this._plugins.has(name)) {
+                debug(`unloading '${name}' plugin`, this.name);
 
-            for (let packetType of handler.Metadata.incomingCapabilities) {
-                this._handlers.delete(packetType);
+                // Unregister packet handlers
+                handler = imports.service.plugins[name];
+
+                for (let type of handler.Metadata.incomingCapabilities) {
+                    this._handlers.delete(type);
+                }
+
+                // Unregister plugin
+                plugin = this._plugins.get(name);
+                plugin.destroy();
+                this._plugins.delete(name);
+                plugin.run_dispose();
             }
-
-            // Unregister plugin
-            this._plugins.get(name).destroy();
-            this._plugins.delete(name);
         } catch (e) {
-            logWarning(`unloading ${name}: ${e.message}`, this.name);
+            logError(e, `${this.name}: unloading ${name}`);
         }
     }
 
-    async _unloadPlugins() {
-        this._plugins.forEach((plugin, name) => this.unloadPlugin(name));
+    async unloadPlugins() {
+        for (let name of this._plugins.keys()) {
+            await this.unloadPlugin(name);
+        }
+    }
+
+    reloadPlugin(name) {
+        this.unloadPlugin(name);
+        this.loadPlugin(name);
+    }
+
+    async reloadPlugins() {
+        for (let name of this.allowed_plugins) {
+            await this.reloadPlugin(name);
+        }
     }
 
     openSettings() {
-        this.service.openSettings(this.id);
+        this.service._preferencesAction(this.id);
     }
 
     destroy() {
         // Close the channel if still connected
-        if (this.connected) {
+        if (this._channel !== null) {
             this._channel.close();
         }
 
         // Synchronously destroy plugins
         this._plugins.forEach(plugin => plugin.destroy());
 
-        // Unexport the GActions and GMenu
+        // Unexport GActions and GMenu
         Gio.DBus.session.unexport_action_group(this._actionsId);
         Gio.DBus.session.unexport_menu_model(this._menuId);
 
-        // Unexport the Device interface
+        // Unexport the Device interface and object
         this._dbus.flush();
         this._dbus_object.remove_interface(this._dbus);
         this._dbus_object.flush();
         this.service.objectManager.unexport(this._dbus_object.g_object_path);
-
-        // Try to avoid any cyclic references from signal handlers
-        GObject.signal_handlers_destroy(this);
-        GObject.signal_handlers_destroy(this.settings);
     }
 });
 
