@@ -5,19 +5,25 @@ const GLib = imports.gi.GLib;
 const GObject = imports.gi.GObject;
 const Gtk = imports.gi.Gtk;
 
-const Contacts = imports.modules.contacts;
-const Lan = imports.service.lan;
 const PluginsBase = imports.service.plugins.base;
 
 
 var Metadata = {
+    label: _('Notifications'),
     id: 'org.gnome.Shell.Extensions.GSConnect.Plugin.Notification',
-    incomingCapabilities: ['kdeconnect.notification', 'kdeconnect.notification.request'],
-    outgoingCapabilities: ['kdeconnect.notification', 'kdeconnect.notification.reply', 'kdeconnect.notification.request'],
+    incomingCapabilities: [
+        'kdeconnect.notification',
+        'kdeconnect.notification.request'
+    ],
+    outgoingCapabilities: [
+        'kdeconnect.notification',
+        // TODO: not supported
+        'kdeconnect.notification.reply',
+        'kdeconnect.notification.request'
+    ],
     actions: {
         withdrawNotification: {
-            summary: _('Cancel Notification'),
-            description: _('Close a local notification remotely by Id'),
+            label: _('Cancel Notification'),
             icon_name: 'preferences-system-notifications-symbolic',
 
             parameter_type: new GLib.VariantType('s'),
@@ -25,8 +31,7 @@ var Metadata = {
             outgoing: ['kdeconnect.notification']
         },
         closeNotification: {
-            summary: _('Close Notification'),
-            description: _('Close a remote notification by Id'),
+            label: _('Close Notification'),
             icon_name: 'preferences-system-notifications-symbolic',
 
             parameter_type: new GLib.VariantType('s'),
@@ -34,8 +39,7 @@ var Metadata = {
             outgoing: ['kdeconnect.notification.request']
         },
         sendNotification: {
-            summary: _('Send Notification'),
-            description: _('Display a local notification remotely'),
+            label: _('Send Notification'),
             icon_name: 'preferences-system-notifications-symbolic',
 
             parameter_type: new GLib.VariantType('a{sv}'),
@@ -61,7 +65,7 @@ var Plugin = GObject.registerClass({
     _init(device) {
         super._init(device, 'notification');
 
-        this.contacts = Contacts.getStore();
+        this.contacts = this.service.contacts;
 
         // Duplicate tracking of telephony notifications
         this._duplicates = new Map();
@@ -82,18 +86,28 @@ var Plugin = GObject.registerClass({
             case 'kdeconnect.notification.request':
                 return this._handleRequest(packet);
 
+            case 'kdeconnect.notification.reply':
+                logWarning('Not implemented', packet.type);
+                return;
+
             default:
                 logWarning('Unknown notification packet', this.device.name);
         }
     }
 
+    connected() {
+        super.connected();
+
+        this.requestNotifications();
+    }
+
     /**
      * Handle an incoming notification or closed report.
      */
     _handleNotification(packet) {
         // A report that a remote notification has been dismissed
         if (packet.body.hasOwnProperty('isCancel')) {
-            this.device.withdraw_notification(packet.body.id);
+            this.device.hideNotification(packet.body.id);
 
         // A remote notification (that hasn't been marked silent)
         } else if (!packet.body.hasOwnProperty('silent')) {
@@ -113,9 +127,7 @@ var Plugin = GObject.registerClass({
         // A request to close a local notification
         //
         // TODO: kdeconnect-android doesn't send these, and will instead send a
-        // kdeconnect.notification packet with isCancel and an id of "0". Other
-        // clients might only support uint32 ids anyways since kdeconnect-kde
-        // only explicitly supports libnotify.
+        // kdeconnect.notification packet with isCancel and an id of "0".
         //
         // For clients that do support it, we report notification ids in the
         // form "type|application-id|notification-id" so we can close it with
@@ -180,29 +192,41 @@ var Plugin = GObject.registerClass({
      * Sending Notifications
      */
     async _uploadIcon(packet, icon) {
-        switch (true) {
-            // TODO: skipping icons for bluetooth connections currently
-            case (this.device.connection_type === 'bluetooth'):
-                this.device.sendPacket(packet);
-                break;
-            case (typeof icon === 'string'):
-                this._uploadNamedIcon(packet, icon);
-                break;
-            case (icon instanceof Gio.BytesIcon):
-                this._uploadBytesIcon(packet, icon.get_bytes());
-                break;
-            case (icon instanceof Gio.FileIcon):
-                this._uploadFileIcon(packet, icon.get_file());
-                break;
-            case (icon instanceof Gio.ThemedIcon):
-                if (icon.hasOwnProperty('name')) {
-                    this._uploadNamedIcon(packet, icon.name);
-                } else {
-                    this._uploadNamedIcon(packet, icon.names[0]);
-                }
-                break;
-            default:
-                this.device.sendPacket(packet);
+        try {
+            // TODO: Currently we skip icons for bluetooth connections
+            if (this.device.connection_type === 'bluetooth') {
+                return this.device.sendPacket(packet);
+            }
+
+            // Normalize icon-name strings into GIcons
+            if (typeof icon === 'string') {
+                icon = new Gio.ThemedIcon({ name: icon });
+            }
+
+            switch (true) {
+                // GBytesIcon
+                case (icon instanceof Gio.BytesIcon):
+                    let bytes = icon.get_bytes();
+                    return this._uploadBytesIcon(packet, bytes);
+                    break;
+
+                // GFileIcon
+                case (icon instanceof Gio.FileIcon):
+                    let file = icon.get_file();
+                    return this._uploadFileIcon(packet, file);
+                    break;
+
+                // GThemedIcon
+                case (icon instanceof Gio.ThemedIcon):
+                    return this._uploadThemedIcon(packet, icon);
+                    break;
+
+                default:
+                    return this.device.sendPacket(packet);
+            }
+        } catch (e) {
+            logError(e);
+            return this.device.sendPacket(packet);
         }
     }
 
@@ -213,58 +237,72 @@ var Plugin = GObject.registerClass({
      * @param {GLib.Bytes} bytes - The themed icon name
      */
     _uploadBytesIcon(packet, bytes) {
-        this._uploadIconStream(
+        return this._uploadIconStream(
             packet,
             Gio.MemoryInputStream.new_from_bytes(bytes),
-            bytes.get_size(),
-            GLib.compute_checksum_for_bytes(
-                GLib.ChecksumType.MD5,
-                bytes.toArray()
-            )
+            bytes.get_size()
         );
     }
 
-    /**
-     * A function for uploading named icons. kdeconnect-android can't handle SVG
-     * icons, so if another is not found the notification will be sent without.
-     *
-     * @param {Core.Packet} packet - The packet for the notification
-     * @param {string} icon_name - The themed icon name
-     */
-    _uploadNamedIcon(packet, icon_name) {
-        let theme = Gtk.IconTheme.get_default();
-        let info = theme.lookup_icon(
-            icon_name,
-            Math.max.apply(null, theme.get_icon_sizes(icon_name)),
-            Gtk.IconLookupFlags.NO_SVG
-        );
-
-        if (info) {
-            this._uploadFileIcon(
-                packet,
-                Gio.File.new_for_path(info.get_filename())
-            );
-        } else {
-            this.device.sendPacket(packet);
-        }
-    }
-
     /**
      * A function for uploading icons as Gio.File objects
      *
      * @param {Core.Packet} packet - The packet for the notification
      * @param {Gio.File} file - A Gio.File object for the icon
      */
-    _uploadFileIcon(packet, file) {
-        this._uploadIconStream(
-            packet,
-            file.read(null),
-            file.query_info('standard::size', 0, null).get_size(),
-            GLib.compute_checksum_for_bytes(
-                GLib.ChecksumType.MD5,
-                file.load_contents(null)[1]
-            )
-        );
+    async _uploadFileIcon(packet, file) {
+        let stream;
+
+        try {
+            stream = await new Promise((resolve, reject) => {
+                file.read_async(GLib.PRIORITY_DEFAULT, null, (file, res) => {
+                    try {
+                        resolve(file.read_finish(res));
+                    } catch (e) {
+                        reject(e);
+                    }
+                });
+            });
+
+            return this._uploadIconStream(
+                packet,
+                stream,
+                file.query_info('standard::size', 0, null).get_size()
+            );
+        } catch (e) {
+            logError(e);
+            this.device.sendPacket(packet);
+        }
+    }
+
+    /**
+     * A function for uploading GThemedIcons
+     *
+     * @param {Core.Packet} packet - The packet for the notification
+     * @param {Gio.ThemedIcon} file - The GIcon to upload
+     */
+    _uploadThemedIcon(packet, icon) {
+        let theme = Gtk.IconTheme.get_default();
+
+        for (let name of icon.names) {
+            // kdeconnect-android doesn't support SVGs so find the largest other
+            let info = theme.lookup_icon(
+                name,
+                Math.max.apply(null, theme.get_icon_sizes(name)),
+                Gtk.IconLookupFlags.NO_SVG
+            );
+
+            // Send the first icon we find from the options
+            if (info) {
+                return this._uploadFileIcon(
+                    packet,
+                    Gio.File.new_for_path(info.get_filename())
+                );
+            }
+        }
+
+        // Fallback to icon-less notification
+        return this.device.sendPacket(packet);
     }
 
     /**
@@ -273,15 +311,12 @@ var Plugin = GObject.registerClass({
      * @param {Core.Packet} packet - The packet for the notification
      * @param {Gio.InputStream} stream - A stream to read the icon bytes from
      * @param {number} size - Size of the icon in bytes
-     * @param {string} checksum - MD5 hash of the icon data
      */
-    async _uploadIconStream(packet, stream, size, checksum) {
-        if (this.device.connection_type === 'tcp') {
-            let transfer = new Lan.Transfer({
-                device: this.device,
-                size: size,
+    async _uploadIconStream(packet, stream, size) {
+        try {
+            let transfer = this.device.createTransfer({
                 input_stream: stream,
-                checksum: checksum
+                size: size
             });
 
             let success = await transfer.upload(packet);
@@ -289,9 +324,8 @@ var Plugin = GObject.registerClass({
             if (!success) {
                 this.device.sendPacket(packet);
             }
-
-        // TODO: skipping icons for bluetooth connections currently
-        } else if (this.device.connection_type === 'bluetooth') {
+        } catch (e) {
+            debug(e);
             this.device.sendPacket(packet);
         }
     }
@@ -301,56 +335,47 @@ var Plugin = GObject.registerClass({
      * See Notification.Listener._sendNotification()
      */
     async sendNotification(notif) {
-        debug(`(${notif.appName}) ${notif.title}: ${notif.text}`);
+        try {
+            debug(`(${notif.appName}) ${notif.title}: ${notif.text}`);
 
-        let applications = JSON.parse(this.settings.get_string('applications'));
+            // TODO: revisit application notification settings
+            let applications = JSON.parse(this.settings.get_string('applications'));
 
-        // New application
-        if (!applications.hasOwnProperty(notif.appName)) {
-            debug(`new application: ${notif.appName}`);
+            // An unknown application
+            if (!applications.hasOwnProperty(notif.appName)) {
+                applications[notif.appName] = {
+                    iconName: 'system-run-symbolic',
+                    enabled: true
+                };
 
-            applications[notif.appName] = {
-                iconName: (typeof notif.icon === 'string') ? notif.icon : 'system-run-symbolic',
-                enabled: true
-            };
-
-            this.settings.set_string(
-                'applications',
-                JSON.stringify(applications)
-            );
-        }
-
-        if (applications[notif.appName].enabled) {
-            let icon = null;
-
-            // Named/Themed Icon
-            if (typeof notif.icon === 'string') {
-                icon = notif.icon;
-                delete notif.icon;
-
-            // Probably a GIcon
-            } else if (typeof notif.icon === 'object') {
-                debug(notif.icon);
-
-                if (notif.icon[0] === 'themed') {
-                    icon = Gio.Icon.deserialize(
-                        new GLib.Variant('(sv)', [
-                            notif.icon[0],
-                            new GLib.Variant('as', notif.icon[1])
-                        ])
-                    );
+                // Only catch icons for strings and GThemedIcon
+                if (typeof notif.icon === 'string') {
+                    applications[notif.appName].iconName = notif.icon;
+                } else if (notif.icon instanceof Gio.ThemedIcon) {
+                    applications[notif.appName].iconName = notif.icon.names[0];
                 }
 
-                delete notif.icon;
+                this.settings.set_string(
+                    'applications',
+                    JSON.stringify(applications)
+                );
             }
 
-            let packet = {
-                id: 0,
-                type: 'kdeconnect.notification',
-                body: notif
-            };
+            // An enabled application
+            if (applications[notif.appName].enabled) {
+                let icon = notif.icon || null;
+                delete notif.icon;
 
-            return await this._uploadIcon(packet, icon);
+                let packet = {
+                    id: 0,
+                    type: 'kdeconnect.notification',
+                    body: notif
+                };
+
+                await this._uploadIcon(packet, icon);
+            }
+        } catch (e) {
+            logError(e);
         }
     }
 
@@ -358,37 +383,49 @@ var Plugin = GObject.registerClass({
      * Receiving Notifications
      */
     async _downloadIcon(packet) {
+        let icon = null;
+        let success, transfer;
+
         try {
             if (!packet.hasOwnProperty('payloadTransferInfo')) {
                 return null;
             }
 
-            let iconStream = Gio.MemoryOutputStream.new_resizable();
-            let icon = null;
-            let success, transfer;
+            let path = GLib.build_filenamev([
+                gsconnect.cachedir,
+                packet.body.payloadHash || `${Date.now()}`
+            ]);
+            let file = Gio.File.new_for_path(path);
+
+            let stream = await new Promise((resolve, reject) => {
+                file.replace_async(null, false, 2, 0, null, (file, res) => {
+                    try {
+                        resolve(file.replace_finish(res));
+                    } catch (e) {
+                        reject(e);
+                    }
+                });
+            });
+
+            let transfer = this.device.createTransfer({
+                output_stream: stream,
+                size: packet.payloadSize
+            });
 
             if (packet.payloadTransferInfo.hasOwnProperty('port')) {
-                transfer = new Lan.Transfer({
-                    device: this.device,
-                    size: packet.payloadSize,
-                    checksum: packet.body.payloadHash,
-                    output_stream: iconStream
-                });
-
                 success = await transfer.download(packet.payloadTransferInfo.port);
 
-            // We ignore bluetooth icon downloads for now
+            // TODO: Skip icons for bluetooth since there's no action to disable
             } else if (packet.payloadTransferInfo.hasOwnProperty('uuid')) {
+                transfer.close();
                 return null;
             }
 
-            iconStream.close(null);
-
             if (success) {
-                icon = new Gio.BytesIcon({ bytes: iconStream.steal_as_bytes() });
+                icon = new Gio.FileIcon({ file: file });
             }
         } catch (e) {
-            logWarning('Failed to download icon', this.device.name);
+            debug(e, this.device.name);
         } finally {
             return icon;
         }
@@ -396,49 +433,49 @@ var Plugin = GObject.registerClass({
 
     /**
      * This mimics _parseEvent() from the telephony plugin, updates the contact
-     * avatar (if necessary), then calls either callNotification() or
-     * smsNotification() from the telephony plugin.
+     * avatar (if necessary), then calls either callNotification() from the
+     * telephony plugin or smsNotification() from the sms plugin.
      *
      * @param {object} notif - The body of a kdeconnect.notification packet
      * @param {object} contact - A contact object
      * @param {Gio.Icon|null} icon - The notification icon (nullable)
      * @param {string} type - The event type; either "missedCall" or "sms"
      */
-    _telephonyNotification(notif, contact, icon, type) {
-        let telephony = imports.service.plugins.telephony.Plugin;
+    async _telephonyNotification(notif, contact, icon, type) {
+        try {
+            // Fabricate a message packet from what we know
+            // TODO: revisit created values
+            let message = {
+                contactName: contact.name,
+                _id: 0,
+                thread_id: 0,
+                address: contact.numbers[0].number,
+                date: parseInt(notif.time),
+                event: type,
+                read: 0,    // Sms.MessageStatus.UNREAD
+                type: 2     // Sms.MessageType.IN
+            };
 
-        // Fabricate a message packet from what we know
-        let message = {
-            contactName: contact.name,
-            _id: 0,
-            thread_id: 0,
-            address: contact.numbers[0].number,
-            date: parseInt(notif.time),
-            event: type,
-            read: 0,    // Sms.MessageStatus.UNREAD
-            type: 2     // Sms.MessageType.IN
-        };
+            // Update contact avatar
+            if (icon instanceof Gio.BytesIcon) {
+                contact = await this.service.contacts.setAvatarPath(
+                    contact.id,
+                    icon.file.get_path()
+                );
+            }
 
-        // Update contact avatar
-        if (!contact.avatar && (icon instanceof Gio.BytesIcon)) {
-            contact.avatar = GLib.build_filenamev([
-                Contacts.CACHE_DIR,
-                GLib.uuid_string_random() + '.jpeg'
-            ]);
-            GLib.file_set_contents(
-                contact.avatar,
-                icon.get_bytes().toArray().toString()
-            );
-            this.contacts._writeCache();
-        }
-
-        if (message.event === 'sms') {
-            message.body = notif.text;
-            telephony.prototype.smsNotification.call(this, contact, message);
-        } else if (message.event === 'missedCall') {
-            // TRANSLATORS: eg. Missed call from John Smith
-            message.body = _('Missed call from %s').format(contact.name);
-            telephony.prototype.callNotification.call(this, contact, message);
+            if (message.event === 'sms') {
+                message.body = notif.text;
+                let sms = imports.service.plugins.sms.Plugin;
+                sms.prototype.smsNotification.call(this, contact, message);
+            } else if (message.event === 'missedCall') {
+                // TRANSLATORS: eg. Missed call from John Smith
+                message.body = _('Missed call from %s').format(contact.name);
+                let telephony = imports.service.plugins.telephony.Plugin;
+                telephony.prototype.callNotification.call(this, contact, message);
+            }
+        } catch (e) {
+            logError(e);
         }
     }
 
@@ -449,99 +486,108 @@ var Plugin = GObject.registerClass({
      * @param {kdeconnect.notification} packet - The notification packet
      */
     async receiveNotification(packet) {
-        //
-        let id = packet.body.id;
-        let body, contact, title;
-        let icon = await this._downloadIcon(packet);
+        try {
+            //
+            let id = packet.body.id;
+            let body, contact, title;
+            let icon = await this._downloadIcon(packet);
 
-        // Check if this is a missed call or SMS notification
-        let isMissedCall = packet.body.id.includes('MissedCall');
-        let isSms = packet.body.id.includes('sms');
+            // Check if this is a sms/telephony notification
+            let isMissedCall = packet.body.id.includes('MissedCall');
+            let isSms = packet.body.id.includes('sms');
 
-        // Check if it's a duplicate early so we can skip unnecessary work
-        let duplicate = this._duplicates.get(packet.body.ticker);
+            // Check if it's been marked as a duplicate
+            let duplicate = this._duplicates.get(packet.body.ticker);
 
-        // This has been marked as a duplicate by the telephony plugin
-        if ((isMissedCall || isSms) && duplicate) {
-            // We've been asked to close this
-            if (duplicate.close) {
-                this.closeNotification(packet.body.id);
-                this._duplicates.delete(packet.body.ticker);
-                return;
-            // We've been asked to silence this, so just track the ID
-            } else if (duplicate.silence) {
-                duplicate.id = packet.body.id;
-                return;
+            if ((isMissedCall || isSms) && duplicate) {
+                // We've been asked to close this
+                if (duplicate.close) {
+                    this.closeNotification(packet.body.id);
+                    this._duplicates.delete(packet.body.ticker);
+                    return;
+                // We've been asked to silence this, so just track the ID
+                } else if (duplicate.silence) {
+                    duplicate.id = packet.body.id;
+                    return;
+                }
             }
-        }
 
-        // If it's a telephony event not marked as a duplicate...
-        if (isMissedCall || isSms) {
-            // Track the id so it can be closed with a telephony notification.
-            this._duplicates.set(packet.body.ticker, { id: packet.body.id });
+            // If it's a telephony event not marked as a duplicate...
+            if (isMissedCall || isSms) {
+                // Track the id so it can be closed with a telephony notification.
+                this._duplicates.set(packet.body.ticker, { id: packet.body.id });
 
-            // Look for a contact with a single phone number, but don't create
-            // one since we only get a number or a name
-            contact = this.contacts.query({
-                name: (isSms) ? packet.body.title : packet.body.text,
-                number: (isSms) ? packet.body.title : packet.body.text,
-                single: true
-            });
-
-            // If found, send this using a telephony plugin method
-            if (contact) {
-                return this._telephonyNotification(
-                    packet.body,
-                    contact,
-                    icon,
-                    (isMissedCall) ? 'missedCall' : 'sms'
-                );
-            }
-        }
-
-        // Emulate a 'missedCall' notification
-        if (isMissedCall) {
-            id = packet.body.ticker;
-            title = packet.body.text;
-            body = _('Missed call from %s').format(packet.body.text);
-        // Emulate an 'sms' notification
-        } else if (isSms) {
-            id = packet.body.ticker;
-            title = packet.body.title;
-            body = packet.body.text;
-        // Ignore 'appName' if it's the same as 'title'
-        } else if (packet.body.appName === packet.body.title) {
-            title = packet.body.title;
-            body = packet.body.text;
-        // Otherwise use the appName as the title
-        } else {
-            title = packet.body.appName;
-            body = packet.body.ticker;
-        }
-
-        // If we don't have a payload icon, fallback on notification type,
-        // appName then device type
-        if (!icon) {
-            if (isMissedCall) {
-                icon = new Gio.ThemedIcon({ name: 'call-missed-symbolic' });
-            } else if (isSms) {
-                icon = new Gio.ThemedIcon({ name: 'sms-symbolic' });
-            } else {
-                icon = new Gio.ThemedIcon({
-                    names: [
-                        packet.body.appName.toLowerCase().replace(' ', '-'),
-                        this.device.symbolic_icon_name
-                    ]
+                // Look for a contact with a single phone number, but don't create
+                // one since we only get a number or a name
+                contact = this.contacts.query({
+                    name: (isSms) ? packet.body.title : packet.body.text,
+                    number: (isSms) ? packet.body.title : packet.body.text
                 });
-            }
-        }
 
-        this.device.showNotification({
-            id: id,
-            title: title,
-            body: body,
-            icon: icon
-        });
+                // If found, send this using a telephony plugin method
+                if (contact) {
+                    return this._telephonyNotification(
+                        packet.body,
+                        contact,
+                        icon,
+                        (isMissedCall) ? 'missedCall' : 'sms'
+                    );
+                }
+            }
+
+            switch (true) {
+                // Emulate a 'missedCall' notification
+                case isMissedCall:
+                    id = packet.body.ticker;
+                    title = packet.body.text;
+                    body = _('Missed call from %s').format(packet.body.text);
+                    break;
+
+                // Emulate an 'sms' notification
+                case isSms:
+                    id = packet.body.ticker;
+                    title = packet.body.title;
+                    body = packet.body.text;
+                    break;
+
+                // Ignore 'appName' if it's the same as 'title'
+                case (packet.body.appName === packet.body.title):
+                    title = packet.body.title;
+                    body = packet.body.text;
+                    break;
+
+                // Otherwise use the appName as the title
+                default:
+                    title = packet.body.appName;
+                    body = packet.body.ticker;
+            }
+
+            // If we don't have a payload icon, fallback on notification type,
+            // appName then device type
+            if (!icon) {
+                if (isMissedCall) {
+                    icon = new Gio.ThemedIcon({ name: 'call-missed-symbolic' });
+                } else if (isSms) {
+                    icon = new Gio.ThemedIcon({ name: 'sms-symbolic' });
+                } else {
+                    icon = new Gio.ThemedIcon({
+                        names: [
+                            packet.body.appName.toLowerCase().replace(' ', '-'),
+                            `${this.device.icon_name}`
+                        ]
+                    });
+                }
+            }
+
+            this.device.showNotification({
+                id: id,
+                title: title,
+                body: body,
+                icon: icon
+            });
+        } catch (e) {
+            logError(e);
+        }
     }
 
     /**
@@ -565,6 +611,7 @@ var Plugin = GObject.registerClass({
 
     /**
      * Close a remote notification.
+     * TODO: ignore local notifications
      *
      * @param {string} id - The remote notification id
      */
