@@ -14,7 +14,6 @@ const PanelMenu = imports.ui.panelMenu;
 const PopupMenu = imports.ui.popupMenu;
 
 const _ = gsconnect._;
-const Color = imports.modules.color;
 const GMenu = imports.shell.gmenu;
 const Tooltip = imports.shell.tooltip;
 
@@ -38,12 +37,24 @@ var Battery = GObject.registerClass({
         this.object = object;
         this.device = device;
 
+        gsconnect.settings.bind(
+            'show-battery',
+            this,
+            'visible',
+            Gio.SettingsBindFlags.GET
+        );
+
         this.label = new St.Label({ text: '' });
         this.add_child(this.label);
 
-        this.icon = new St.Icon({ icon_size: 16 });
+        this.icon = new St.Icon();
         this.add_child(this.icon);
 
+        this._deviceId = this.device.connect(
+            'g-properties-changed',
+            this.update.bind(this)
+        );
+
         // Battery proxy
         this.battery = this.object.get_interface(BATTERY_INTERFACE);
 
@@ -54,40 +65,60 @@ var Battery = GObject.registerClass({
             );
         }
 
-        this.object.connect('interface-added', (obj, iface) => {
-            if (iface.g_interface_name === BATTERY_INTERFACE) {
-                this.battery = iface;
-                this._batteryId = this.battery.connect(
-                    'g-properties-changed',
-                    this.update.bind(this)
-                );
-            }
-        });
+        this._interfaceAddedId = this.object.connect(
+            'interface-added',
+            this._onInterfaceAdded.bind(this)
+        );
 
-        this.object.connect('interface-removed', (obj, iface) => {
-            if (iface.g_interface_name === BATTERY_INTERFACE) {
-                this.battery = iface;
-                this.battery.disconnect(this._batteryId);
-                delete this._batteryId;
-                delete this.battery;
-            }
-        });
+        this._interfaceRemovedId = this.object.connect(
+            'interface-removed',
+            this._onInterfaceRemoved.bind(this)
+        );
+
+        this.connect('notify::mapped', this.update.bind(this));
 
         // Cleanup
-        this.connect('destroy', (actor) => {
-            if (actor._batteryId && actor.battery) {
-                actor.battery.disconnect(actor._batteryId);
-            }
-        });
+        this.connect('destroy', this._onDestroy);
+    }
+
+    _onDestroy(actor) {
+        actor.device.disconnect(actor._deviceId);
+
+        if (actor._batteryId && actor.battery) {
+            actor.battery.disconnect(actor._batteryId);
+        }
+
+        actor.object.disconnect(actor._interfaceAddedId);
+        actor.object.disconnect(actor._interfaceRemovedId);
+    }
+
+    _onInterfaceAdded(object, iface) {
+        if (iface.g_interface_name === BATTERY_INTERFACE) {
+            this.battery = iface;
+            this._batteryId = this.battery.connect(
+                'g-properties-changed',
+                this.update.bind(this)
+            );
+        }
+    }
+
+    _onInterfaceRemoved(object, iface) {
+        if (iface.g_interface_name === BATTERY_INTERFACE) {
+            iface.disconnect(this._batteryId);
+            this.battery = null;
+            this._batteryId = 0;
+        }
     }
 
     update(battery) {
         if (!this.mapped) { return; }
 
-        this.icon.visible = (this.battery && this.battery.Level > -1);
-        this.label.visible = (this.battery && this.battery.Level > -1);
+        let connected = (this.device.Connected && this.device.Paired);
+        let visible = (connected && this.battery && this.battery.Level > -1);
+        this.icon.visible = visible;
+        this.label.visible = visible;
 
-        if (!this.icon.visible || !this.label.visible) { return; }
+        if (!visible) { return; }
 
         this.icon.icon_name = this.battery.IconName;
         this.label.text = this.battery.Level + '%';
@@ -99,13 +130,11 @@ var Battery = GObject.registerClass({
  * A Device Icon
  */
 var Icon = GObject.registerClass({
-    GtypeName: 'GSConnectShellDeviceIcon'
+    GTypeName: 'GSConnectShellDeviceIcon'
 }, class Icon extends St.DrawingArea {
 
     _init(object, device) {
         super._init({
-            width: 48,
-            height: 48,
             reactive: true,
             track_hover: true,
             y_align: Clutter.ActorAlign.START,
@@ -115,33 +144,32 @@ var Icon = GObject.registerClass({
         this.object = object;
         this.device = device;
 
+        // Watch for scale changes
+        this._themeContext = St.ThemeContext.get_for_stage(global.stage);
+        this._themeContextId = this._themeContext.connect(
+            'notify::scale-factor',
+            this._onThemeChanged.bind(this)
+        );
+
+        // Watch for icon theme changes
+        this._iconTheme = Gtk.IconTheme.get_default();
+        this._iconThemeId = this._iconTheme.connect(
+            'changed',
+            this._onThemeChanged.bind(this)
+        );
+
         this.tooltip = new Tooltip.Tooltip({
             parent: this,
             markup: this.device.Name,
-            y_offset: 16
+            y_offset: 16 * this._themeContext.scale_factor
         });
 
-        // Device Type
-        this._theme = Gtk.IconTheme.get_default();
-        this.icon = this._theme.load_surface(
-            this.device.IconName,
-            32,
-            1,
-            null,
-            Gtk.IconLookupFlags.FORCE_SIZE
+        // Device Status
+        this._propertiesId = device.connect(
+            'g-properties-changed',
+            () => this.queue_repaint()
         );
 
-        this._themeSignal = this._theme.connect('changed', () => {
-            this.icon = this._theme.load_surface(
-                this.device.IconName,
-                32,
-                1,
-                null,
-                Gtk.IconLookupFlags.FORCE_SIZE
-            );
-            this.queue_repaint();
-        });
-
         // Battery proxy
         this.battery = this.object.get_interface(BATTERY_INTERFACE);
 
@@ -152,74 +180,150 @@ var Icon = GObject.registerClass({
             );
         }
 
-        this.object.connect('interface-added', (obj, iface) => {
-            if (iface.g_interface_name === BATTERY_INTERFACE) {
-                this.battery = iface;
-                this._batteryId = this.battery.connect(
-                    'g-properties-changed',
-                    () => this.queue_repaint()
-                );
-            }
-        });
+        // Watch for interface changes
+        this._interfaceAddedId = this.object.connect(
+            'interface-added',
+            this._onInterfaceAdded.bind(this)
+        );
 
-        this.object.connect('interface-removed', (obj, iface) => {
-            if (iface.g_interface_name === BATTERY_INTERFACE) {
-                this.battery = iface;
-                this.battery.disconnect(this._batteryId);
-                delete this._batteryId;
-                delete this.battery;
-            }
-        });
+        this._interfaceRemovedId = this.object.connect(
+            'interface-removed',
+            this._onInterfaceRemoved.bind(this)
+        );
 
-        // Device Status
-        this._propertiesId = device.connect(
-            'g-properties-changed',
+        // Redraw when mapped
+        this._mappedId = this.connect(
+            'notify::mapped',
             () => this.queue_repaint()
         );
 
         // Cleanup
-        this.connect('destroy', (actor) => {
-            actor.device.disconnect(actor._propertiesId);
-            actor._theme.disconnect(actor._themeSignal);
+        this.connect('destroy', this._onDestroy);
+    }
 
-            if (actor._batteryId && actor.battery) {
-                actor.battery.disconnect(actor._batteryId);
-            }
-        });
+    _onDestroy(actor) {
+        actor.disconnect(actor._mappedId);
+
+        actor.device.disconnect(actor._propertiesId);
+        actor._iconTheme.disconnect(actor._iconThemeId);
+        actor._themeContext.disconnect(actor._themeContextId);
+
+        if (actor._batteryId && actor.battery) {
+            actor.battery.disconnect(actor._batteryId);
+        }
+
+        actor.object.disconnect(actor._interfaceAddedId);
+        actor.object.disconnect(actor._interfaceRemovedId);
+    }
+
+    _onInterfaceAdded(object, iface) {
+        if (iface.g_interface_name === BATTERY_INTERFACE) {
+            this.battery = iface;
+            this._batteryId = iface.connect(
+                'g-properties-changed',
+                () => this.queue_repaint()
+            );
+        }
+    }
+
+    _onInterfaceRemoved(object, iface) {
+        if (iface.g_interface_name === BATTERY_INTERFACE) {
+            iface.disconnect(this._batteryId);
+            this.battery = null;
+        }
+    }
+
+    _update() {
+        let scale = this._themeContext.scale_factor;
+
+        // Get the unscaled size as defined in CSS
+        if (this._width === undefined) {
+            [this._width, this._height] = this.get_size();
+            this._width = this._width / scale;
+            this._height = this._height / scale;
+        }
+
+        // Resize the actor if the scale has changed
+        if (!this._scale || this._scale !== scale) {
+            this._scale = scale;
+            this.tooltip.y_offset = 16 * scale;
+            this.width = this._width * scale;
+            this.height = this._height * scale;
+        }
+
+        // Size icon 2/3 of total, scaled and rounded up to multiple of 8
+        let size = Math.ceil((this._width / 1.5 * scale) / 8) * 8;
+        this._offset = size / 2;
+
+        this._icon = this._iconTheme.load_surface(
+            this.device.IconName,
+            size,
+            scale,
+            null,
+            Gtk.IconLookupFlags.FORCE_SIZE
+        );
+    }
+
+    _onThemeChanged() {
+        // Only update after we've been added to the stage
+        if (this.width) {
+            this._update();
+            this.queue_repaint();
+        }
     }
 
     get battery_color() {
-        return Color.hsv2rgb(
-            this.battery.Level / 100 * 120,
-            100,
-            100 - (this.battery.Level / 100 * 15)
-        );
+        let h = (this.battery.Level / 100 * 120) / 360;
+        let s = 1;
+        let v = (100 - (this.battery.Level / 100 * 15)) / 100;
+
+        let i = Math.floor(h * 6);
+        let f = h * 6 - i;
+        let p = v * (1 - s);
+        let q = v * (1 - f * s);
+        let t = v * (1 - (1 - f) * s);
+
+        let r, g, b;
+
+        switch (i % 6) {
+            case 0: r = v, g = t, b = p; break;
+            case 1: r = q, g = v, b = p; break;
+            case 2: r = p, g = v, b = t; break;
+            case 3: r = p, g = q, b = v; break;
+            case 4: r = t, g = p, b = v; break;
+            case 5: r = v, g = p, b = q; break;
+        }
+
+        return [r, g, b];
     }
 
     get battery_label() {
         let { Charging, Level, Time } = this.battery;
 
         if (Level === 100) {
-            // TRANSLATORS: Fully Charged
+            // TRANSLATORS: When the battery level is 100%
             return _('Fully Charged');
         } else if (Time === 0) {
-            // TRANSLATORS: <percentage> (Estimating…)
+            // TRANSLATORS: When no time estimate for the battery is available
+            // EXAMPLE: 42% (Estimating…)
             return _('%d%% (Estimating…)').format(Level);
         }
 
         Time = Time / 60;
-        let minutes = Time % 60;
+        let minutes = Math.floor(Time % 60);
         let hours = Math.floor(Time / 60);
 
         if (Charging) {
-            // TRANSLATORS: <percentage> (<hours>:<minutes> Until Full)
+            // TRANSLATORS: Estimated time until battery is charged
+            // EXAMPLE: 42% (1:15 Until Full)
             return _('%d%% (%d\u2236%02d Until Full)').format(
                 Level,
                 hours,
                 minutes
             );
         } else {
-            // TRANSLATORS: <percentage> (<hours>:<minutes> Remaining)
+            // TRANSLATORS: Estimated time until battery is empty
+            // EXAMPLE: 42% (12:15 Remaining)
             return _('%d%% (%d\u2236%02d Remaining)').format(
                 Level,
                 hours,
@@ -229,28 +333,35 @@ var Icon = GObject.registerClass({
     }
 
     vfunc_repaint() {
-        if (!this.visible) { return; }
+        if (!this.mapped) { return; }
 
-        let [width, height] = this.get_surface_size();
+        if (this._icon === undefined) {
+            this._update();
+        }
+
+        let cr = this.get_context();
+
+        // Dimensions have already been scaled
+        let [width, height] = this.get_size();
         let xc = width / 2;
         let yc = height / 2;
 
-        let cr = this.get_context();
-        let thickness = 3;
+        // Colored circle width
+        let thickness = Math.ceil(this.width / 16);
         let r = Math.min(xc, yc) - (thickness / 2);
         cr.setLineWidth(thickness);
 
         // Icon
-        cr.setSourceSurface(this.icon, xc - 16, yc - 16);
+        cr.setSourceSurface(this._icon, xc - this._offset, yc - this._offset);
         cr.paint();
 
         if (!this.device.Connected) {
             cr.setOperator(Cairo.Operator.HSL_SATURATION);
             cr.setSourceRGB(0, 0, 0);
-            cr.maskSurface(this.icon, xc - 16, yc - 16);
+            cr.maskSurface(this._icon, xc - this._offset, yc - this._offset);
             cr.fill();
 
-            this.tooltip.markup = _('Reconnect <b>%s</b>').format(this.device.Name);
+            this.tooltip.markup = _('Reconnect');
             this.tooltip.icon_name = 'view-refresh-symbolic';
 
             cr.setSourceRGB(0.8, 0.8, 0.8);
@@ -260,9 +371,7 @@ var Icon = GObject.registerClass({
             cr.arc(xc, yc, r, 1.48 * Math.PI, 1.47 * Math.PI);
             cr.stroke();
         } else if (!this.device.Paired) {
-            // TRANSLATORS: eg. Pair <b>Google Pixel</b>
-            this.tooltip.markup = _('Pair <b>%s</b>').format(this.device.Name) +
-                                 '\n\n' + this.device.EncryptionInfo;
+            this.tooltip.markup = _('Pair') + '\n\n' + this.device.EncryptionInfo;
             this.tooltip.icon_name = 'channel-insecure-symbolic';
 
             cr.setSourceRGB(0.95, 0.0, 0.0);
@@ -304,8 +413,8 @@ var Icon = GObject.registerClass({
                 cr.fill();
             }
         } else {
-            this.tooltip.markup = _('Configure <b>%s</b>').format(this.device.Name);
-            this.tooltip.icon_name = 'preferences-other-symbolic';
+            this.tooltip.markup = _('Settings');
+            this.tooltip.icon_name = 'preferences-system-symbolic';
             cr.setSourceRGB(0.8, 0.8, 0.8);
             cr.arc(xc, yc, r, 0, 2 * Math.PI);
             cr.stroke();
@@ -338,7 +447,7 @@ var Button = GObject.registerClass({
         this.object = object;
         this.device = device;
 
-        this.connect('clicked', this._onClicked.bind(this));
+        this.connect('clicked', this._onClicked);
     }
 
     _onClicked(button) {
@@ -370,17 +479,19 @@ var Menu = class Menu extends PopupMenu.PopupMenuSection {
         this.deviceBox = new PopupMenu.PopupBaseMenuItem({
             can_focus: false,
             reactive: false,
-            style_class: 'popup-menu-item gsconnect-device-box'
+            style_class: 'popup-menu-item gsconnect-device'
         });
         this.deviceBox.actor.remove_child(this.deviceBox._ornamentLabel);
         this.deviceBox.actor.vertical = false;
         this.addMenuItem(this.deviceBox);
 
+        // Device Icon with Battery Circle
         this.deviceButton = new Button(object, iface);
         this.deviceBox.actor.add_child(this.deviceButton);
 
+        // Title Bar & Plugin/Status Bar
         this.controlBox = new St.BoxLayout({
-            style_class: 'gsconnect-control-box',
+            style_class: 'gsconnect-device-box',
             vertical: true,
             x_expand: true
         });
@@ -388,7 +499,7 @@ var Menu = class Menu extends PopupMenu.PopupMenuSection {
 
         // Title Bar
         this.titleBar = new St.BoxLayout({
-            style_class: 'gsconnect-title-bar'
+            style_class: 'gsconnect-device-title'
         });
         this.controlBox.add_child(this.titleBar);
 
@@ -401,7 +512,7 @@ var Menu = class Menu extends PopupMenu.PopupMenuSection {
 
         // Title Bar -> Separator
         let nameSeparator = new St.Widget({
-            style_class: 'popup-separator-menu-item gsconnect-title-separator',
+            style_class: 'popup-separator-menu-item',
             x_expand: true,
             y_expand: true,
             y_align: Clutter.ActorAlign.CENTER
@@ -416,17 +527,13 @@ var Menu = class Menu extends PopupMenu.PopupMenuSection {
         this.pluginBar = new GMenu.FlowBox({
             action_group: iface.action_group,
             menu_model: iface.menu_model,
-            style_class: 'gsconnect-plugin-bar'
+            style_class: 'gsconnect-device-actions'
         });
-        this.pluginBar.connect(
-            'submenu-toggle',
-            this._onSubmenuToggle.bind(this)
-        );
         this.controlBox.add_child(this.pluginBar);
 
         // Status Bar
         this.statusBar = new St.BoxLayout({
-            style_class: 'gsconnect-status-bar',
+            style_class: 'gsconnect-device-status',
             y_align: Clutter.ActorAlign.FILL,
             y_expand: true
         });
@@ -438,19 +545,19 @@ var Menu = class Menu extends PopupMenu.PopupMenuSection {
         });
         this.statusBar.add_child(this.statusLabel);
 
-        // Hide the submenu when the device menu is closed
-        this._getTopMenu().connect('open-state-changed', (actor, open) => {
-            if (!open && this._submenu !== undefined) {
-                this.box.remove_child(this._submenu.actor);
-                this._submenu = undefined;
-            }
-        });
-
-        // Watch GSettings & Properties
-        this._gsettingsId = gsconnect.settings.connect(
-            'changed',
-            this._sync.bind(this)
+        //
+        this.pluginBar.connect(
+            'notify::submenu',
+            this._onSubmenuChanged.bind(this)
         );
+
+        // Hide the submenu when the device menu is closed
+        this._getTopMenu().connect(
+            'open-state-changed',
+            this._onOpenStateChanged.bind(this)
+        );
+
+        // Watch Properties
         this._propertiesId = this.device.connect(
             'g-properties-changed',
             this._sync.bind(this)
@@ -461,43 +568,38 @@ var Menu = class Menu extends PopupMenu.PopupMenuSection {
         );
     }
 
-    _onSubmenuToggle(flowbox, button) {
-        // Close (remove) any currently opened submenu
-        if (this._submenu !== undefined) {
-            this.box.remove_child(this._submenu.actor);
-        }
-
-        // Open (add) the submenu if it's a new menu...
-        if (this._submenu !== button.submenu) {
-            this._submenu = button.submenu;
-            this.box.add_child(this._submenu.actor);
-        // ...otherwise unset the current submenu
-        } else {
-            this._submenu = undefined;
+    _onOpenStateChanged(actor, open) {
+        if (!open && this._submenu !== undefined) {
+            this.pluginBar.submenu = undefined;
         }
     }
 
-    _sync() {
-        debug(`${this.device.Name} (${this.device.Id})`);
+    _onSubmenuChanged(flowbox) {
+        // Close (remove) any currently opened submenu
+        if (this._submenu !== undefined) {
+            this.box.remove_child(this._submenu);
+        }
 
+        this._submenu = flowbox.submenu;
+
+        // Open (add) the submenu if it's a new menu...
+        if (this._submenu !== undefined) {
+            this.box.add_child(this._submenu);
+        }
+    }
+
+    _sync(proxy, changed, invalidated) {
         if (!this.actor.mapped) { return; }
 
-        let { Connected, Paired } = this.device;
+        debug(`${this.device.Name} menu`);
 
         // Title Bar
         this.nameLabel.text = this.device.Name;
 
-        // TODO: might as well move this to actors.js
-        if (Connected && Paired && gsconnect.settings.get_boolean('show-battery')) {
-            this.deviceBattery.visible = true;
-            this.deviceBattery.update();
-        } else {
-            this.deviceBattery.visible = false;
-        }
-
         // Plugin/Status Bar visibility
+        let { Connected, Paired } = this.device;
         this.pluginBar.visible = (Connected && Paired);
-        this.statusBar.visible = (!Connected || !Paired);
+        this.statusBar.visible = !this.pluginBar.visible;
 
         if (!Connected) {
             this.statusLabel.text = _('Device is disconnected');
@@ -509,7 +611,6 @@ var Menu = class Menu extends PopupMenu.PopupMenuSection {
     destroy() {
         this.actor.disconnect(this._mappedId);
         this.device.disconnect(this._propertiesId);
-        gsconnect.settings.disconnect(this._gsettingsId);
 
         super.destroy();
     }
@@ -537,14 +638,20 @@ var Indicator = class Indicator extends PanelMenu.Button {
         this.menu.addMenuItem(this.deviceMenu);
 
         // Watch GSettings & Properties
-        this._gsettingsId = gsconnect.settings.connect('changed', this._sync.bind(this));
-        this._propertiesId = this.device.connect('g-properties-changed', this._sync.bind(this));
+        this._gsettingsId = gsconnect.settings.connect(
+            'changed',
+            this._sync.bind(this)
+        );
+        this._propertiesId = this.device.connect(
+            'g-properties-changed',
+            this._sync.bind(this)
+        );
 
         this._sync();
     }
 
     get symbolic_icon() {
-        let icon_name = this.device.SymbolicIconName;
+        let icon_name = `${this.device.IconName}-symbolic`;
 
         if (!this.device.Paired) {
             let rgba = new Gdk.RGBA({ red: 0.95, green: 0, blue: 0, alpha: 0.9 });
@@ -555,8 +662,8 @@ var Indicator = class Indicator extends PanelMenu.Button {
         return new Gio.ThemedIcon({ name: icon_name });
     }
 
-    _sync() {
-        debug(`${this.device.Name} (${this.device.Id})`);
+    async _sync() {
+        debug(`${this.device.Name} indicator`);
 
         let { Connected, Paired } = this.device;
 
